import json
import uuid
from collections.abc import MutableSequence
from inspect import isasyncgenfunction
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import parse_qsl, quote_plus

from blacksheep.plugins import json as json_plugin

from .exceptions cimport MessageAborted


cdef class Content:

    def __init__(self,
                 bytes content_type,
                 bytes data):
        self.type = content_type
        self.body = data
        self.length = len(data)

    async def read(self):
        return self.body


cdef class StreamedContent(Content):

    def __init__(self,
                 bytes content_type,
                 object data_provider):
        self.type = content_type
        self.body = None
        self.length = -1
        self.generator = data_provider
        if not isasyncgenfunction(data_provider):
            raise ValueError('Data provider must be an async generator')

    async def read(self):
        value = bytearray()

        async for chunk in self.generator():
            value.extend(chunk)

        self.body = bytes(value)
        self.length = len(self.body)
        return self.body

    async def get_parts(self):
        async for chunk in self.generator():
            yield chunk


cdef class ASGIContent(Content):

    def __init__(self, object receive):
        self.type = None
        self.body = None
        self.length = -1
        self.receive = receive

    cpdef void dispose(self):
        self.receive = None
        self.body = None

    async def stream(self):
        while True:
            message = await self.receive()

            if message.get('type') == 'http.disconnect':
                raise MessageAborted()

            yield message.get('body', b'')

            if not message.get('more_body'):
                break

        yield b''

    async def read(self):
        if self.body is not None:
            return self.body
        value = bytearray()

        while True:
            message = await self.receive()

            if message.get('type') == 'http.disconnect':
                raise MessageAborted()

            value.extend(message.get('body', b''))

            if not message.get('more_body'):
                break

        self.body = bytes(value)
        self.length = len(self.body)
        return self.body


cdef class TextContent(Content):

    def __init__(self, str text):
        super().__init__(b'text/plain; charset=utf-8', text.encode('utf8'))


cdef class HTMLContent(Content):

    def __init__(self, str html):
        super().__init__(b'text/html; charset=utf-8', html.encode('utf8'))


cdef class JSONContent(Content):

    def __init__(self, object data, dumps=json_plugin.dumps):
        super().__init__(b'application/json', dumps(data).encode('utf8'))


cdef dict parse_www_form_urlencoded(str content):
    # application/x-www-form-urlencoded
    cdef str key, value
    cdef dict data = {}
    for key, value in parse_qsl(content):
        if key in data:
            if isinstance(data[key], str):
                data[key] = [data[key], value]
            else:
                data[key].append(value)
        else:
            data[key] = value
    return data


cpdef dict parse_www_form(str content):
    return parse_www_form_urlencoded(content)


cdef object try_decode(bytes value, str encoding):
    try:
        return value.decode(encoding or 'utf8')
    except:
        return value


cdef dict multiparts_to_dictionary(list parts):
    cdef str key
    cdef str charset
    cdef data = {}
    cdef FormPart part

    for part in parts:
        key = part.name.decode('utf8')
        if part.charset:
            charset = part.charset.encode()
        else:
            charset = None

        # NB: we cannot assume that the value of a multipart form part can be decoded as UTF8;
        # here we try to decode it, just to be more consistent with values read from www-urlencoded form data
        if part.file_name:
            # Files need special handling, must be kept as-is
            if key in data:
                data[key].append(part)
            else:
                data[key] = [part]
        else:
            if key in data:
                if isinstance(data[key], list):
                    data[key].append(try_decode(part.data, charset))
                else:
                    data[key] = [data[key], try_decode(part.data, charset)]
            else:
                data[key] = try_decode(part.data, charset)

    return data


cpdef void write_multipart_part(FormPart part, bytearray destination):
    # https://tools.ietf.org/html/rfc7578#page-4
    destination.extend(b'Content-Disposition: form-data; name="')
    destination.extend(part.name)
    destination.extend(b'"')
    if part.file_name:
        destination.extend(b'; filename="')
        destination.extend(part.file_name)
        destination.extend(b'"\r\n')
    if part.content_type:
        destination.extend(b'Content-Type: ')
        destination.extend(part.content_type)
    destination.extend(b'\r\n\r\n')
    destination.extend(part.data)
    destination.extend(b'\r\n')


cpdef bytes write_www_form_urlencoded(data: Union[dict, list]):
    # application/x-www-form-urlencoded
    if isinstance(data, list):
        values = data
    else:
        values = data.items()

    cdef list contents = []

    for key, value in values:
        if isinstance(value, MutableSequence):
            for item in value:
                contents.append(quote_plus(key) + '=' + quote_plus(str(item)))
        else:
            contents.append(quote_plus(key) + '=' + quote_plus(str(value)))
    return ('&'.join(contents)).encode('utf8')


cdef class FormContent(Content):

    def __init__(self, data: Union[Dict[str, str], List[Tuple[str, str]]]):
        super().__init__(b'application/x-www-form-urlencoded', write_www_form_urlencoded(data))


cdef class FormPart:

    def __init__(self,
                 bytes name,
                 bytes data,
                 bytes content_type: Optional[bytes]=None,
                 bytes file_name: Optional[bytes]=None,
                 bytes charset: Optional[bytes] = None):
        self.name = name
        self.data = data
        self.file_name = file_name
        self.content_type = content_type
        self.charset = charset

    def __eq__(self, other):
        if isinstance(other, FormPart):
            return other.name == self.name and other.file_name == self.file_name and other.content_type == self.content_type and other.charset == self.charset and other.data == self.data
        if other is None:
            return False
        return NotImplemented

    def __repr__(self):
        return f'<FormPart {self.name} - at {id(self)}>'


cdef class MultiPartFormData(Content):

    def __init__(self, list parts):
        self.parts = parts
        self.boundary = b'------' + str(uuid.uuid4()).replace('-', '').encode()
        super().__init__(b'multipart/form-data; boundary=' + self.boundary, write_multipart_form_data(self))


cpdef bytes write_multipart_form_data(MultiPartFormData data):
    cdef bytearray contents = bytearray()
    cdef FormPart part
    for part in data.parts:
        contents.extend(b'--')
        contents.extend(data.boundary)
        contents.extend(b'\r\n')
        write_multipart_part(part, contents)
    contents.extend(b'--')
    contents.extend(data.boundary)
    contents.extend(b'--\r\n')
    return bytes(contents)


HtmlContent = HTMLContent  # for backward compatibility
JsonContent = JSONContent  # for backward compatibility



import json
from typing import Any, Dict, Optional

import jwt
import pkg_resources
import pytest
from guardpost.authentication import Identity
from guardpost.authorization import AuthorizationContext, Policy, UnauthorizedError
from guardpost.common import AuthenticatedRequirement
from guardpost.jwks import JWKS, InMemoryKeysProvider, KeysProvider
from pytest import raises

from blacksheep.server.application import Application
from blacksheep.server.authentication import (
    AuthenticateChallenge,
    AuthenticationHandler,
)
from blacksheep.server.authentication.jwt import JWTBearerAuthentication
from blacksheep.server.authorization import (
    AuthorizationWithoutAuthenticationError,
    Requirement,
    allow_anonymous,
    auth,
    get_www_authenticated_header_from_generic_unauthorized_error,
)
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.test_files_serving import get_folder_path


def get_file_path(file_name, folder_name: str = "res") -> str:
    return pkg_resources.resource_filename(__name__, f"./{folder_name}/{file_name}")


# region JWTBearer


def test_jwt_bearer_authentication_sets_authority_as_issuer():
    authority = "https://sts.windows.net/a2884dee-52e8-4034-8ce2-6b48e18d1ae7/"
    auth = JWTBearerAuthentication(
        valid_audiences=["api://0ed1cebe-b7ca-45c5-a4bf-a8d586c18d31"],
        authority=authority,
    )
    assert auth._validator._valid_issuers == [authority]


def test_jwt_bearer_authentication_throws_for_missing_issuer():
    with pytest.raises(TypeError):
        JWTBearerAuthentication(valid_audiences=["foo"])


def get_test_jwks() -> JWKS:
    with open(get_file_path("jwks.json"), mode="rt", encoding="utf8") as jwks_file:
        jwks_dict = json.loads(jwks_file.read())
    return JWKS.from_dict(jwks_dict)


@pytest.fixture(scope="session")
def default_keys_provider() -> KeysProvider:
    return InMemoryKeysProvider(get_test_jwks())


def get_access_token(
    kid: str, payload: Dict[str, Any], *, fake_kid: Optional[str] = None
):
    with open(get_file_path(f"{kid}.pem"), "r") as key_file:
        private_key = key_file.read()

    return jwt.encode(
        payload,
        private_key,
        algorithm="RS256",
        headers={"kid": kid if not fake_kid else fake_kid},
    )


# endregion


class MockAuthHandler(AuthenticationHandler):
    def __init__(self, identity=None):
        if identity is None:
            identity = Identity({"id": "001", "name": "Charlie Brown"}, "JWT")
        self.identity = identity

    async def authenticate(self, context: Any) -> Optional[Identity]:
        context.identity = self.identity
        return context.identity


class MockNotAuthHandler(AuthenticationHandler):
    async def authenticate(self, context: Any) -> Optional[Identity]:
        context.identity = Identity({"id": "007"})
        # NB: an identity without authentication scheme is treated
        # as anonymous identity
        return context.identity


class AccessTokenCrashingHandler(AuthenticationHandler):
    async def authenticate(self, context: Any) -> Optional[Identity]:
        raise AuthenticateChallenge(
            "Bearer",
            None,
            {
                "error": "Invalid access token",
                "error_description": "Access token expired",
            },
        )


class AdminRequirement(Requirement):
    def handle(self, context: AuthorizationContext):
        identity = context.identity

        if identity is not None and identity["role"] == "admin":
            context.succeed(self)


class AdminsPolicy(Policy):
    def __init__(self):
        super().__init__("admin", AdminRequirement())


@pytest.mark.asyncio
async def test_authentication_sets_identity_in_request(app):
    app.use_authentication().add(MockAuthHandler())

    identity = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.identity
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204

    assert identity is not None
    assert identity["id"] == "001"
    assert identity["name"] == "Charlie Brown"


@pytest.mark.asyncio
async def test_authorization_unauthorized_error(app):
    app.use_authentication().add(MockAuthHandler())

    app.use_authorization().add(AdminsPolicy())

    @auth("admin")
    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401


@pytest.mark.asyncio
async def test_authorization_policy_success(app):
    admin = Identity({"id": "001", "name": "Charlie Brown", "role": "admin"}, "JWT")

    app.use_authentication().add(MockAuthHandler(admin))

    app.use_authorization().add(AdminsPolicy())

    @auth("admin")
    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204


@pytest.mark.asyncio
async def test_authorization_default_allows_anonymous(app):
    app.use_authentication().add(MockAuthHandler())

    app.use_authorization().add(AdminsPolicy())

    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204


@pytest.mark.asyncio
async def test_authorization_supports_default_require_authenticated(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401


@pytest.mark.asyncio
async def test_static_files_allow_anonymous_by_default(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.serve_files(get_folder_path("files"))

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/lorem-ipsum.txt"), MockReceive(), MockSend())

    assert app.response.status == 200
    content = await app.response.text()
    assert content == "Lorem ipsum dolor sit amet\n"


@pytest.mark.asyncio
async def test_static_files_support_authentication(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.serve_files(get_folder_path("files"), allow_anonymous=False)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/lorem-ipsum.txt"), MockReceive(), MockSend())

    assert app.response.status == 401


@pytest.mark.asyncio
async def test_static_files_support_authentication_by_route(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.serve_files(get_folder_path("files"), allow_anonymous=False)
    app.serve_files(get_folder_path("files2"), allow_anonymous=True, root_path="/login")

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/lorem-ipsum.txt"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/login/index.html"), MockReceive(), MockSend())

    assert app.response.status == 200
    content = await app.response.text()
    assert (
        content
        == """<!DOCTYPE html>
<html>
  <head>
    <title>Example.</title>
    <link rel="stylesheet" type="text/css" href="/styles/main.css" />
  </head>
  <body>
    <h1>Lorem ipsum</h1>
    <p>Dolor sit amet.</p>
    <script src="/scripts/main.js"></script>
  </body>
</html>
"""
    )


@pytest.mark.asyncio
async def test_authorization_supports_allow_anonymous(app):
    from blacksheep.server.responses import text

    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @allow_anonymous()
    @app.router.get("/")
    async def home():
        return text("Hi There!")

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200


@pytest.mark.asyncio
async def test_authentication_challenge_response(app):
    app.use_authentication().add(AccessTokenCrashingHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401
    header = app.response.get_single_header(b"WWW-Authenticate")

    assert header is not None
    assert header == (
        b'Bearer, error="Invalid access token", '
        b'error_description="Access token expired"'
    )


@pytest.mark.asyncio
async def test_authorization_strategy_without_authentication_raises(app):
    app.use_authorization()

    with raises(AuthorizationWithoutAuthenticationError):
        app.prepare()


@pytest.mark.parametrize(
    "scheme,realm,parameters,expected_value",
    [
        ["Basic", None, None, b"Basic"],
        ["Bearer", "Mushrooms Kingdom", None, b'Bearer realm="Mushrooms Kingdom"'],
        [
            "Bearer",
            "Mushrooms Kingdom",
            {
                "title": "Something",
                "error": "Invalid access token",
                "error_description": "access token expired",
            },
            b'Bearer realm="Mushrooms Kingdom", '
            b'title="Something", '
            b'error="Invalid access token", '
            b'error_description="access token expired"',
        ],
    ],
)
def test_authentication_challenge_error(scheme, realm, parameters, expected_value):
    error = AuthenticateChallenge(scheme, realm, parameters)

    header = error.get_header()
    assert header[0] == b"WWW-Authenticate"
    assert header[1] == expected_value


@pytest.mark.parametrize(
    "exc,expected_value",
    [
        (UnauthorizedError(None, [], scheme="AAD"), b"AAD"),
        (UnauthorizedError(None, [], scheme="example"), b"example"),
        (
            UnauthorizedError(None, [], scheme="Something-Something"),
            b"Something-Something",
        ),
    ],
)
def test_get_www_authenticated_header_from_generic_unauthorized_error(
    exc, expected_value
):
    header = get_www_authenticated_header_from_generic_unauthorized_error(exc)

    assert header is not None
    name, value = header
    assert name == b"WWW-Authenticate"
    assert value == expected_value


@pytest.mark.asyncio
async def test_authorization_default_requires_authenticated_user(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization()

    @app.router.get("/")
    async def home():
        return None

    @auth()
    @app.router.get("/admin")
    async def admin():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response.status == 204

    await app(get_example_scope("GET", "/admin"), MockReceive(), MockSend())
    assert app.response.status == 401


@pytest.mark.asyncio
async def test_jwt_bearer_authentication(app, default_keys_provider):
    app.use_authentication().add(
        JWTBearerAuthentication(
            valid_audiences=["a"],
            valid_issuers=["b"],
            keys_provider=default_keys_provider,
        )
    )

    identity: Optional[Identity] = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.identity
        return None

    await app.start()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity.is_authenticated() is False

    # request with valid Bearer Token
    access_token = get_access_token(
        "0",
        {
            "aud": "a",
            "iss": "b",
            "id": "001",
            "name": "Charlie Brown",
        },
    )
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", b"Bearer " + access_token.encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity["id"] == "001"
    assert identity["name"] == "Charlie Brown"

    # request with invalid Bearer Token (invalid audience)
    access_token = get_access_token(
        "0",
        {
            "aud": "NO",
            "iss": "b",
            "id": "001",
            "name": "Charlie Brown",
        },
    )
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", b"Bearer " + access_token.encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity.is_authenticated() is False

    # request with invalid Bearer Token (invalid header)
    access_token = get_access_token(
        "0",
        {
            "aud": "a",
            "iss": "b",
            "id": "001",
            "name": "Charlie Brown",
        },
    )
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", access_token.encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity.is_authenticated() is False


def test_set_authentication_strategy_more_than_once(app: Application):
    auth_strategy = app.use_authentication()
    assert app.use_authentication() is auth_strategy


def test_set_authorization_strategy_more_than_once(app: Application):
    auth_strategy = app.use_authorization()
    assert app.use_authorization() is auth_strategy



"""
Root module of the framework. This module re-exports the most commonly
used types to reduce the verbosity of the imports statements.
"""
__author__ = "Roberto Prevato <roberto.prevato@gmail.com>"

from .contents import Content as Content
from .contents import FormContent as FormContent
from .contents import FormPart as FormPart
from .contents import HTMLContent as HTMLContent
from .contents import HtmlContent as HtmlContent
from .contents import JSONContent as JSONContent
from .contents import MultiPartFormData as MultiPartFormData
from .contents import StreamedContent as StreamedContent
from .contents import TextContent as TextContent
from .contents import parse_www_form as parse_www_form
from .cookies import Cookie as Cookie
from .cookies import CookieSameSiteMode as CookieSameSiteMode
from .cookies import datetime_from_cookie_format as datetime_from_cookie_format
from .cookies import datetime_to_cookie_format as datetime_to_cookie_format
from .cookies import parse_cookie as parse_cookie
from .exceptions import HTTPException as HTTPException
from .headers import Header as Header
from .headers import Headers as Headers
from .messages import Message as Message
from .messages import Request as Request
from .messages import Response as Response
from .server.application import Application as Application
from .server.authorization import allow_anonymous as allow_anonymous
from .server.authorization import auth as auth
from .server.bindings import ClientInfo as ClientInfo
from .server.bindings import FromBytes as FromBytes
from .server.bindings import FromCookie as FromCookie
from .server.bindings import FromFiles as FromFiles
from .server.bindings import FromForm as FromForm
from .server.bindings import FromHeader as FromHeader
from .server.bindings import FromJSON as FromJSON
from .server.bindings import FromQuery as FromQuery
from .server.bindings import FromRoute as FromRoute
from .server.bindings import FromServices as FromServices
from .server.bindings import FromText as FromText
from .server.bindings import ServerInfo as ServerInfo
from .server.responses import ContentDispositionType as ContentDispositionType
from .server.responses import FileInput as FileInput
from .server.responses import accepted as accepted
from .server.responses import bad_request as bad_request
from .server.responses import created as created
from .server.responses import file as file
from .server.responses import forbidden as forbidden
from .server.responses import html as html
from .server.responses import json as json
from .server.responses import moved_permanently as moved_permanently
from .server.responses import no_content as no_content
from .server.responses import not_found as not_found
from .server.responses import not_modified as not_modified
from .server.responses import ok as ok
from .server.responses import permanent_redirect as permanent_redirect
from .server.responses import pretty_json as pretty_json
from .server.responses import redirect as redirect
from .server.responses import see_other as see_other
from .server.responses import status_code as status_code
from .server.responses import temporary_redirect as temporary_redirect
from .server.responses import text as text
from .server.responses import unauthorized as unauthorized
from .server.routing import Route as Route
from .server.routing import RouteException as RouteException
from .server.routing import Router as Router
from .server.routing import RoutesRegistry as RoutesRegistry
from .server.templating import use_templates as use_templates
from .server.websocket import WebSocket as WebSocket
from .server.websocket import WebSocketDisconnectError as WebSocketDisconnectError
from .server.websocket import WebSocketError as WebSocketError
from .server.websocket import WebSocketState as WebSocketState
from .url import URL as URL
from .url import InvalidURL as InvalidURL



from dataclasses import asdict, is_dataclass
from functools import lru_cache
from typing import Any

from jinja2 import Environment, PackageLoader, Template, select_autoescape
from rodi import Container

from blacksheep import Content, Response


@lru_cache(1200)
def template_name(name: str):
    if not name.endswith(".html"):
        return name + ".html"
    return name


def get_response(html: str):
    return Response(200, [(b"Cache-Control", b"no-cache")]).with_content(
        Content(b"text/html; charset=utf-8", html.encode("utf8"))
    )


def render_template(template: Template, *args, **kwargs):
    return template.render(*args, **kwargs)


async def render_template_async(template: Template, *args, **kwargs):
    return await template.render_async(*args, **kwargs)


def use_templates(app, loader: PackageLoader, enable_async: bool = False):
    env = getattr(app, "jinja_environment", None)
    if not env:
        env = Environment(
            loader=loader,
            autoescape=select_autoescape(["html", "xml"]),
            auto_reload=app.debug,
            enable_async=enable_async,
        )

        app.jinja_environment = env
        app.templates_environment = env

        if isinstance(app.services, Container):

            def get_jinja_env() -> Environment:
                return env

            app.services.add_singleton_by_factory(get_jinja_env)
            app.services.add_alias("jinja_environment", Environment)
            app.services.add_alias("jinja", Environment)
            app.services.add_alias("templates_environment", Environment)
        else:
            raise TypeError(
                "Application services must be an instance of `rodi.Container`."
            )
        env.globals["app"] = app

    if enable_async:

        async def async_view(name: str, model: Any = None, **kwargs):
            return get_response(
                await render_template_async(
                    env.get_template(template_name(name)), model, **kwargs
                )
            )

        return async_view

    def sync_view(name: str, model: Any = None, **kwargs):
        return get_response(
            render_template(env.get_template(template_name(name)), model, **kwargs)
        )

    return sync_view


def model_to_view_params(model):
    if isinstance(model, dict):
        return model
    if is_dataclass(model):
        return asdict(model)
    if hasattr(model, "__dict__"):
        return model.__dict__
    return model


def view(
    jinja_environment: Environment, name: str, model: Any = None, **kwargs
) -> Response:
    """
    Returns a Response object with HTML obtained from synchronous rendering.

    Use this when `enable_async` is set to False when calling `use_templates`.
    """
    if model:
        return get_response(
            render_template(
                jinja_environment.get_template(template_name(name)),
                **model_to_view_params(model),
                **kwargs
            )
        )
    return get_response(
        render_template(jinja_environment.get_template(template_name(name)), **kwargs)
    )


async def view_async(
    jinja_environment: Environment, name: str, model: Any = None, **kwargs
) -> Response:
    """
    Returns a Response object with HTML obtained from synchronous rendering.
    """
    if model:
        return get_response(
            await render_template_async(
                jinja_environment.get_template(template_name(name)),
                **model_to_view_params(model),
                **kwargs
            )
        )
    return get_response(
        await render_template_async(
            jinja_environment.get_template(template_name(name)), **kwargs
        )
    )



import time

import pytest
from cryptography.fernet import Fernet

from blacksheep.cookies import parse_cookie
from blacksheep.messages import Request
from blacksheep.server.responses import text
from blacksheep.sessions import JSONSerializer, Session, SessionMiddleware
from blacksheep.sessions.crypto import FernetEncryptor
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend


def test_friendly_exception_for_request_without_session():
    request = Request("GET", b"/", None)

    with pytest.raises(TypeError):
        request.session


def test_session_base_methods():
    session = Session()

    assert "foo" not in session

    session["foo"] = "lorem ipsum"

    assert "foo" in session
    assert session["foo"] == "lorem ipsum"

    del session["foo"]

    assert "foo" not in session

    session.set("foo", "lorem ipsum")
    assert session.get("foo") == "lorem ipsum"

    assert session.get("ufo", ...) is ...

    session.update({"a": 1, "b": 2, "c": 3})

    assert session["a"] == 1
    assert session["b"] == 2
    assert session["c"] == 3


@pytest.mark.parametrize(
    "values,expected_len",
    [
        [
            {},
            0,
        ],
        [
            {
                "a": 1,
                "b": 2,
            },
            2,
        ],
        [
            {
                "a": 1,
                "b": 2,
                "c": 3,
            },
            3,
        ],
        [{"a": 1, "b": 2, "c": 3, "d": 4}, 4],
    ],
)
def test_session_length(values, expected_len):
    session = Session(values)
    assert len(session) == expected_len


def test_session_clear():
    session = Session({"a": 1, "b": 2, "c": 3, "d": 4})

    session.clear()
    assert session.modified
    assert len(session) == 0


def test_session_to_dict():
    value = {"a": 1, "b": 2, "c": 3, "d": 4}
    session = Session(value)

    assert session.to_dict() == value
    assert session.to_dict() is not value


@pytest.mark.parametrize(
    "values",
    [
        {},
        {"a": 1},
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
        {"a": 1, "b": 2, "c": 3, "d": 4},
    ],
)
def test_session_equality(values):
    session = Session(values)

    assert session == session
    assert session == Session(values)
    assert session == values

    values["x"] = True

    assert session != Session(values)
    assert session != values


def test_session_inequality():
    session = Session()
    assert (session == []) is False
    assert (session == "") is False


def test_session_modified():
    session = Session()

    assert session.modified is False

    session["foo"] = "lorem ipsum"

    assert session.modified is True

    session = Session({"foo": "lorem ipsum"})

    assert session.modified is False

    # any set item marks the session as modified,
    # it doesn't matter if the end values are the same
    session["foo"] = "lorem ipsum"

    assert session.modified is True


def test_session_key_error():
    session = Session()

    with pytest.raises(KeyError):
        session["foo"]


@pytest.mark.parametrize(
    "value,session",
    [
        ["{}", Session()],
        ['{"lorem":"ipsum"}', Session({"lorem": "ipsum"})],
        ['{"lorem":"ipsum ✨"}', Session({"lorem": "ipsum ✨"})],
    ],
)
def test_session_json_serializer(value, session):
    serializer = JSONSerializer()

    assert serializer.write(session) == value
    assert serializer.read(value) == session


@pytest.mark.asyncio
async def test_session_middleware_basics(app):
    app.middlewares.append(SessionMiddleware("LOREM_IPSUM"))

    @app.router.get("/")
    def home(request: Request):
        session = request.session

        assert isinstance(session, Session)
        session["foo"] = "Some value"

        return text("Hello, World")

    @app.router.get("/second")
    def second(request: Request):
        session = request.session

        assert "foo" in session
        assert session["foo"] == "Some value"

        return text("Hello, World")

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_single(b"Set-Cookie")
    assert session_set_cookie is not None

    cookie = parse_cookie(session_set_cookie)

    await app(
        get_example_scope("GET", "/second", {"cookie": f"session={cookie.value}"}),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_first(b"Set-Cookie")
    assert session_set_cookie is None


@pytest.mark.asyncio
async def test_session_middleware_use_method(app):
    app.use_sessions("LOREM_IPSUM")

    @app.router.get("/")
    def home(request: Request):
        session = request.session

        assert isinstance(session, Session)
        session["foo"] = "Some value"

        return text("Hello, World")

    @app.router.get("/second")
    def second(request: Request):
        session = request.session

        assert "foo" in session
        assert session["foo"] == "Some value"

        return text("Hello, World")

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_single(b"Set-Cookie")
    assert session_set_cookie is not None

    cookie = parse_cookie(session_set_cookie)

    await app(
        get_example_scope("GET", "/second", {"cookie": f"session={cookie.value}"}),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_first(b"Set-Cookie")
    assert session_set_cookie is None


@pytest.mark.asyncio
async def test_session_middleware_with_encryptor(app):
    app.middlewares.append(
        SessionMiddleware(
            "LOREM_IPSUM", encryptor=FernetEncryptor(Fernet.generate_key())
        )
    )

    @app.router.get("/")
    def home(request: Request):
        session = request.session

        assert isinstance(session, Session)
        session["foo"] = "Some value"

        return text("Hello, World")

    @app.router.get("/second")
    def second(request: Request):
        session = request.session

        assert "foo" in session
        assert session["foo"] == "Some value"

        return text("Hello, World")

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_single(b"Set-Cookie")
    assert session_set_cookie is not None

    cookie = parse_cookie(session_set_cookie)

    await app(
        get_example_scope("GET", "/second", {"cookie": f"session={cookie.value}"}),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_first(b"Set-Cookie")
    assert session_set_cookie is None


@pytest.mark.asyncio
async def test_session_middleware_handling_of_invalid_signature(app):
    app.middlewares.append(SessionMiddleware("LOREM_IPSUM"))

    @app.router.get("/")
    def home(request: Request):
        session = request.session

        assert isinstance(session, Session)
        assert len(session) == 0
        assert "user_id" not in session

        return text("Hello, World")

    await app.start()

    # arrange invalid session cookie
    impostor_middleware = SessionMiddleware("DOLOR_SIT_AMET")

    forged_cookie = impostor_middleware.write_session(Session({"user_id": "hahaha"}))

    await app(
        get_example_scope("GET", "/", {"cookie": f"session={forged_cookie}"}),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200


@pytest.mark.asyncio
async def test_session_middleware_handling_of_expired_signature(app):
    app.middlewares.append(SessionMiddleware("LOREM_IPSUM", session_max_age=1))

    @app.router.get("/")
    def home(request: Request):
        session = request.session

        assert isinstance(session, Session)
        session["foo"] = "Some value"

        return text("Hello, World")

    @app.router.get("/second")
    def second(request: Request):
        session = request.session

        assert "foo" not in session

        return text("Hello, World")

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_single(b"Set-Cookie")
    assert session_set_cookie is not None

    cookie = parse_cookie(session_set_cookie)

    time.sleep(2)

    await app(
        get_example_scope("GET", "/second", {"cookie": f"session={cookie.value}"}),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    session_set_cookie = response.headers.get_first(b"Set-Cookie")
    assert session_set_cookie is None


@pytest.mark.asyncio
async def test_session_middleware_handling_of_invalid_encrypted_signature(app):
    app.middlewares.append(
        SessionMiddleware(
            "LOREM_IPSUM", encryptor=FernetEncryptor(Fernet.generate_key())
        )
    )

    @app.router.get("/")
    def home(request: Request):
        session = request.session

        assert isinstance(session, Session)
        assert len(session) == 0
        assert "user_id" not in session

        return text("Hello, World")

    await app.start()

    # arrange invalid session cookie
    impostor_middleware = SessionMiddleware(
        "LOREM_IPSUM", encryptor=FernetEncryptor(Fernet.generate_key())
    )

    forged_cookie = impostor_middleware.write_session(Session({"user_id": "hahaha"}))

    await app(
        get_example_scope("GET", "/", {"cookie": f"session={forged_cookie}"}),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200


def test_exception_for_invalid_max_age():
    with pytest.raises(ValueError):
        SessionMiddleware("example", session_max_age=0)

    with pytest.raises(ValueError):
        SessionMiddleware("example", session_max_age=-10)



from datetime import datetime

import pytest

from blacksheep import (
    Cookie,
    CookieSameSiteMode,
    datetime_from_cookie_format,
    datetime_to_cookie_format,
    parse_cookie,
    scribe,
)
from blacksheep.cookies import CookieValueExceedsMaximumLength

COOKIES = [
    (
        "Foo",
        "Power",
        None,
        None,
        None,
        False,
        False,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo=Power",
    ),
    (
        "Foo",
        "Hello World;",
        None,
        None,
        None,
        False,
        False,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo=Hello%20World%3B",
    ),
    (
        "Foo; foo",
        "Hello World;",
        None,
        None,
        None,
        False,
        False,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo%3B%20foo=Hello%20World%3B",
    ),
    (
        "Foo",
        "Power",
        datetime(2018, 8, 17, 20, 55, 4),
        None,
        None,
        False,
        False,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo=Power; Expires=Fri, 17 Aug 2018 20:55:04 GMT",
    ),
    (
        "Foo",
        "Power",
        datetime(2018, 8, 17, 20, 55, 4),
        "something.org",
        None,
        False,
        False,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo=Power; Expires=Fri, 17 Aug 2018 20:55:04 GMT; Domain=something.org",
    ),
    (
        "Foo",
        "Power",
        datetime(2018, 8, 17, 20, 55, 4),
        "something.org",
        "/",
        True,
        False,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo=Power; Expires=Fri, 17 Aug 2018 20:55:04 GMT; Domain=something.org; Path=/; HttpOnly",
    ),
    (
        "Foo",
        "Power",
        datetime(2018, 8, 17, 20, 55, 4),
        "something.org",
        "/",
        True,
        True,
        -1,
        CookieSameSiteMode.UNDEFINED,
        b"Foo=Power; Expires=Fri, 17 Aug 2018 20:55:04 GMT; Domain=something.org; Path=/; HttpOnly; Secure",
    ),
    (
        "Foo",
        "Power",
        datetime(2018, 8, 17, 20, 55, 4),
        "something.org",
        "/",
        True,
        True,
        -1,
        CookieSameSiteMode.LAX,
        b"Foo=Power; Expires=Fri, 17 Aug 2018 20:55:04 GMT; Domain=something.org; Path=/; HttpOnly; Secure; SameSite=Lax",
    ),
    (
        "Foo",
        "Power",
        datetime(2018, 8, 17, 20, 55, 4),
        "something.org",
        "/",
        True,
        True,
        200,
        CookieSameSiteMode.STRICT,
        b"Foo=Power; Expires=Fri, 17 Aug 2018 20:55:04 GMT; Max-Age=200; "
        b"Domain=something.org; Path=/; HttpOnly; Secure; SameSite=Strict",
    ),
]


@pytest.mark.parametrize(
    "name,value,expires,domain,path,http_only,secure,max_age,same_site,expected_result",
    COOKIES,
)
def test_write_cookie(
    name,
    value,
    expires,
    domain,
    path,
    http_only,
    secure,
    max_age,
    same_site,
    expected_result,
):
    cookie = Cookie(
        name,
        value,
        expires,
        domain,
        path,
        http_only,
        secure,
        max_age,
        same_site,
    )
    value = scribe.write_response_cookie(cookie)
    assert value == expected_result


@pytest.mark.parametrize(
    "name,value,expires,domain,path,http_only,secure,max_age,same_site,expected_result",
    COOKIES,
)
def test_parse_cookie(
    name,
    value,
    expires,
    domain,
    path,
    http_only,
    secure,
    max_age,
    same_site,
    expected_result,
):
    cookie = parse_cookie(expected_result)
    assert cookie is not None
    assert cookie.name == name
    assert cookie.value == value
    if expires:
        assert cookie.expires is not None
        assert cookie.expires == expires
    assert cookie.domain == domain
    assert cookie.path == path
    assert cookie.http_only == http_only
    assert cookie.secure == secure
    if max_age is not None:
        assert cookie.max_age == max_age


@pytest.mark.parametrize(
    "value,expected_result",
    [
        [b"Sun, 27-Jan-2019 20:40:54 GMT", datetime(2019, 1, 27, 20, 40, 54)],
        [b"Sun, 27 Jan 2019 20:40:54 GMT", datetime(2019, 1, 27, 20, 40, 54)],
        [b"Wed, 21 Oct 2015 07:28:00 GMT", datetime(2015, 10, 21, 7, 28, 00)],
        [b"Thu, 31-Dec-37 23:55:55 GMT", datetime(2037, 12, 31, 23, 55, 55)],
        [b"Tuesday, 08-Feb-94 14:15:29 GMT", datetime(1994, 2, 8, 14, 15, 29)],
        [b"09 Feb 1994 22:23:32 GMT", datetime(1994, 2, 9, 22, 23, 32)],
        [b"08-Feb-94 14:15:29 GMT", datetime(1994, 2, 8, 14, 15, 29)],
        [b"08-Feb-1994 14:15:29 GMT", datetime(1994, 2, 8, 14, 15, 29)],
    ],
)
def test_datetime_from_cookie_format(value, expected_result):
    parsed = datetime_from_cookie_format(value)
    assert parsed == expected_result


@pytest.mark.parametrize(
    "expected_result,value",
    [
        [b"Sun, 27 Jan 2019 20:40:54 GMT", datetime(2019, 1, 27, 20, 40, 54)],
        [b"Wed, 21 Oct 2015 07:28:00 GMT", datetime(2015, 10, 21, 7, 28, 00)],
    ],
)
def test_datetime_to_cookie_format(expected_result, value):
    bytes_value = datetime_to_cookie_format(value)
    assert bytes_value == expected_result


@pytest.mark.parametrize(
    "value,expected_name,expected_value,expected_path",
    [
        (
            b"ARRAffinity=c12038089a7sdlkj1237192873; Path=/; HttpOnly; Domain=example.scm.azurewebsites.net",
            "ARRAffinity",
            "c12038089a7sdlkj1237192873",
            "/",
        ),
        (
            b"ARRAffinity=c12038089a7sdlkj1237192873;Path=/;HttpOnly;Domain=example.scm.azurewebsites.net",
            "ARRAffinity",
            "c12038089a7sdlkj1237192873",
            "/",
        ),
        (
            b"1P_JAR=2020-08-23-11; expires=Tue, 22-Sep-2020 11:13:40 GMT; path=/; domain=.google.com; Secure",
            "1P_JAR",
            "2020-08-23-11",
            "/",
        ),
        (
            b"NID=204=0K7PurlER1icDcU_vBBCFWff0gPjtSX3saNz-AXBmkjWGi7RWl_XEeV4uAUuHdX0qsAJbaAhl8E-fZTjwMlTyB9Du_bkal2PHdlnz6h0iKsBNjC5ee8JePM-0PW6hCKdyxyORH6Dzhd7kkvJBhZzk6HQz0QeP8vi9h9eDGL0RGs; expires=Mon, 22-Feb-2021 11:13:40 GMT; path=/; domain=.google.com; HttpOnly",
            "NID",
            "204=0K7PurlER1icDcU_vBBCFWff0gPjtSX3saNz-AXBmkjWGi7RWl_XEeV4uAUuHdX0qsAJbaAhl8E-fZTjwMlTyB9Du_bkal2PHdlnz6h0iKsBNjC5ee8JePM-0PW6hCKdyxyORH6Dzhd7kkvJBhZzk6HQz0QeP8vi9h9eDGL0RGs",
            "/",
        ),
        (
            b"session=gAAAAABgVeIWAXQ5iCbIgXThcx9IFORha534yIqw2ZjnqiTKIw7xBcnk-Tc8pvuTpLEuFSv3NRJkr83WBdhc0dpjZrEGBUNCFV8YK17hka43KanCxW5FMhrP00AxvGYyKZ2-vy4CEUIcsN92JAvV763u_ZCZzSpraw==",
            "session",
            "gAAAAABgVeIWAXQ5iCbIgXThcx9IFORha534yIqw2ZjnqiTKIw7xBcnk-Tc8pvuTpLEuFSv3NRJkr83WBdhc0dpjZrEGBUNCFV8YK17hka43KanCxW5FMhrP00AxvGYyKZ2-vy4CEUIcsN92JAvV763u_ZCZzSpraw==",
            None,
        ),
    ],
)
def test_parse_cookie_separators(value, expected_name, expected_value, expected_path):
    cookie = parse_cookie(value)

    assert cookie is not None
    assert cookie.name == expected_name
    assert cookie.value == expected_value
    assert cookie.path == expected_path


def test_raise_for_value_exceeding_length():
    with pytest.raises(CookieValueExceedsMaximumLength):
        Cookie("crash", "A" * 4967)

    with pytest.raises(CookieValueExceedsMaximumLength):
        Cookie("crash", "A" * 5000)



from blacksheep.contents import ASGIContent
from blacksheep.messages import Request


def get_request_url_from_scope(
    scope,
    base_path: str = "",
    include_query: bool = True,
    trailing_slash: bool = False,
) -> str:
    """
    Function used for diagnostic reasons, for example to generate URLs in pages with
    detailed information about internal server errors.

    Do not use this method for logic that must generate full request URL, since it
    doesn't handle Forward and X-Forwarded* headers - use instead:

    > from blacksheep.messages import get_absolute_url_to_path, get_request_absolute_url
    """
    try:
        path = scope["path"]
        protocol = scope["scheme"]
        host, port = scope["server"]
    except KeyError as key_error:
        raise ValueError(f"Invalid scope: {key_error}")

    if protocol == "http" and port == 80:
        port_part = ""
    elif protocol == "https" and port == 443:
        port_part = ""
    else:
        port_part = f":{port}"

    if trailing_slash:
        path = path + "/"

    query_part = (
        ""
        if not include_query or not scope.get("query_string")
        else ("?" + scope.get("query_string").decode("utf8"))
    )
    return f"{protocol}://{host}{port_part}{base_path}{path}{query_part}"


def get_request_url(request: Request) -> str:
    """
    Function used for diagnostic reasons, for example to generate URLs in pages with
    detailed information about internal server errors.

    Do not use this method for logic that must generate full request URL, since it
    doesn't handle Forward and X-Forwarded* headers - use instead:

    > from blacksheep.messages import get_absolute_url_to_path, get_request_absolute_url
    """
    return get_request_url_from_scope(request.scope)


def incoming_request(scope, receive=None) -> Request:
    """
    Function used to simulate incoming requests from an ASGI scope.
    This function is intentionally not used in
    `blacksheep.server.application.Application`.
    """
    request = Request.incoming(
        scope["method"],
        scope["raw_path"],
        scope["query_string"],
        list(scope["headers"]),
    )
    request.scope = scope
    if receive:
        request.content = ASGIContent(receive)
    return request


def get_full_path(scope) -> bytes:
    """
    Returns the full path of the HTTP message from an ASGI scope.
    """
    path = scope["path"].encode()
    query = scope["query_string"]

    if query:
        return path + b"?" + query

    return path



import sys
from dataclasses import dataclass
from inspect import Parameter, _ParameterKind
from typing import List, Optional, Sequence, Union

import pytest
from guardpost.authentication import Identity, User
from pytest import raises
from rodi import Container, Services, inject

from blacksheep import Request
from blacksheep.server.bindings import (
    Binder,
    ExactBinder,
    FromHeader,
    FromJSON,
    FromQuery,
    FromRoute,
    FromServices,
    HeaderBinder,
    IdentityBinder,
    JSONBinder,
    QueryBinder,
    RequestBinder,
    RouteBinder,
    ServiceBinder,
)
from blacksheep.server.normalization import (
    AmbiguousMethodSignatureError,
    NormalizationError,
    RouteBinderMismatch,
    UnsupportedSignatureError,
    _check_union,
    _get_raw_bound_value_type,
    get_binders,
    normalize_handler,
    normalize_middleware,
)
from blacksheep.server.routing import Route


class Pet:
    def __init__(self, name):
        self.name = name


class Cat(Pet):
    ...


class Dog(Pet):
    ...


@dataclass
class SomeService:
    pass


def test_parameters_get_binders_default_query():
    def handler(a, b, c):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[1].parameter_name == "b"
    assert binders[2].parameter_name == "c"


@pytest.mark.parametrize(
    "annotation_type", [Identity, User, Optional[User], Optional[Identity]]
)
def test_identity_binder_by_param_type(annotation_type):
    async def handler(param):
        ...

    handler.__annotations__["param"] = annotation_type

    binders = get_binders(Route(b"/", handler), Services())

    assert isinstance(binders[0], IdentityBinder)


def test_parameters_get_binders_from_route():
    def handler(a, b, c):
        ...

    binders = get_binders(Route(b"/:a/:b/:c", handler), Services())

    assert all(isinstance(binder, RouteBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[1].parameter_name == "b"
    assert binders[2].parameter_name == "c"


def test_parameters_get_binders_from_services_by_name():
    def handler(a, b, c):
        ...

    binders = get_binders(
        Route(b"/", handler), {"a": object(), "b": object(), "c": object()}
    )

    assert all(isinstance(binder, ServiceBinder) for binder in binders)
    assert binders[0].expected_type == "a"
    assert binders[1].expected_type == "b"
    assert binders[2].expected_type == "c"


def test_parameters_get_binders_from_services_by_type():
    def handler(a: str, b: int, c: Cat):
        ...

    binders = get_binders(
        Route(b"/", handler), {str: object(), int: object(), Cat: object()}
    )

    assert all(isinstance(binder, ServiceBinder) for binder in binders)
    assert binders[0].expected_type is str
    assert binders[1].expected_type is int
    assert binders[2].expected_type is Cat


def test_parameters_get_binders_from_body():
    def handler(a: Cat):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    assert len(binders) == 1
    binder = binders[0]

    assert isinstance(binder, JSONBinder)
    assert binder.expected_type is Cat
    assert binder.required is True


def test_parameters_get_binders_from_body_optional():
    def handler(a: Optional[Cat]):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    assert len(binders) == 1
    binder = binders[0]

    assert isinstance(binder, JSONBinder)
    assert binder.expected_type is Cat
    assert binder.required is False


def test_parameters_get_binders_simple_types_default_from_query():
    def handler(a: str, b: int, c: bool):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == str
    assert binders[1].parameter_name == "b"
    assert binders[1].expected_type == int
    assert binders[2].parameter_name == "c"
    assert binders[2].expected_type == bool


def test_parameters_get_binders_list_types_default_from_query():
    def handler(a: List[str], b: List[int], c: List[bool]):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == List[str]
    assert binders[1].parameter_name == "b"
    assert binders[1].expected_type == List[int]
    assert binders[2].parameter_name == "c"
    assert binders[2].expected_type == List[bool]


def test_parameters_get_binders_list_types_default_from_query_optional():
    def handler(a: Optional[List[str]]):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert all(binder.required is False for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == List[str]


def test_parameters_get_binders_list_types_default_from_query_required():
    def handler(a: List[str]):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert all(binder.required for binder in binders)


def test_parameters_get_binders_sequence_types_default_from_query():
    def handler(a: Sequence[str], b: Sequence[int], c: Sequence[bool]):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == Sequence[str]
    assert binders[1].parameter_name == "b"
    assert binders[1].expected_type == Sequence[int]
    assert binders[2].parameter_name == "c"
    assert binders[2].expected_type == Sequence[bool]


def test_throw_for_ambiguous_binder_multiple_from_body():
    def handler(a: Cat, b: Dog):
        ...

    with pytest.raises(AmbiguousMethodSignatureError):
        get_binders(Route(b"/", handler), Services())


def test_does_not_throw_for_forward_ref():
    def handler(a: "Cat"):
        ...

    get_binders(Route(b"/", handler), Services())

    def handler(a: List["str"]):
        ...

    get_binders(Route(b"/", handler), Services())

    def handler(a: Optional[List["Cat"]]):
        ...

    get_binders(Route(b"/", handler), Services())

    def handler(a: FromQuery["str"]):
        ...

    get_binders(Route(b"/", handler), Services())


def test_combination_of_sources():
    def handler(
        a: FromQuery[List[str]],
        b: FromServices[Dog],
        c: FromJSON[Cat],
        d: FromRoute[str],
        e: FromHeader[str],
    ):
        ...

    binders = get_binders(Route(b"/:d", handler), {Dog: Dog("Snoopy")})

    assert isinstance(binders[0], QueryBinder)
    assert isinstance(binders[1], ServiceBinder)
    assert isinstance(binders[2], JSONBinder)
    assert isinstance(binders[3], RouteBinder)
    assert isinstance(binders[4], HeaderBinder)
    assert binders[0].parameter_name == "a"
    assert binders[1].parameter_name == "b"
    assert binders[2].parameter_name == "c"
    assert binders[3].parameter_name == "d"
    assert binders[4].parameter_name == "e"


def test_implicit_from_services_only_when_annotation_is_none():
    def handler(dog):
        ...

    binders = get_binders(Route(b"/", handler), {"dog": Dog("Snoopy")})

    assert isinstance(binders[0], ServiceBinder)

    def handler(dog: str):
        ...

    binders = get_binders(Route(b"/", handler), {"dog": Dog("Snoopy")})

    assert isinstance(binders[0], QueryBinder)


def test_from_query_specific_name():
    class FromExampleQuery(FromQuery[str]):
        name = "example"

    @inject()
    def handler(a: FromExampleQuery):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is str
    assert binder.required is True
    assert binder.parameter_name == "example"


def test_from_query_unspecified_type():
    def handler(a: FromQuery):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is List[str]
    assert binder.required is True
    assert binder.parameter_name == "a"


def test_from_query_optional_type():
    def handler(a: FromQuery[Optional[str]]):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is str
    assert binder.required is False
    assert binder.parameter_name == "a"


def test_from_query_optional_type_with_union():
    def handler(a: FromQuery[Union[None, str]]):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is str
    assert binder.required is False
    assert binder.parameter_name == "a"


def test_check_union():
    optional, value = _check_union(
        Parameter("foo", kind=_ParameterKind.POSITIONAL_ONLY),
        Union[None, str],
        len,
    )
    assert optional is True
    assert value is str


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_check_union_or_none():
    optional, value = _check_union(
        Parameter("foo", kind=_ParameterKind.POSITIONAL_ONLY),
        str | None,
        len,
    )
    assert optional is True
    assert value is str


def test_from_query_optional_list_type():
    def handler(a: FromQuery[Optional[List[str]]]):
        ...

    binders = get_binders(Route(b"/", handler), Services())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is List[str]
    assert binder.required is False
    assert binder.parameter_name == "a"


def test_from_header_specific_name():
    class FromExampleHeader(FromHeader[str]):
        name = "example"

    @inject()
    def handler(a: FromExampleHeader):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert isinstance(binders[0], HeaderBinder)
    assert binders[0].expected_type is str
    assert binders[0].parameter_name == "example"


def test_from_header_accept_language_example():
    class AcceptLanguageHeader(FromHeader[str]):
        name = "accept-language"

    @inject()
    def handler(a: AcceptLanguageHeader):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert isinstance(binders[0], HeaderBinder)
    assert binders[0].expected_type is str
    assert binders[0].parameter_name == "accept-language"


def test_raises_for_route_mismatch():
    def handler(a: FromRoute[str]):
        ...

    with raises(RouteBinderMismatch):
        get_binders(Route(b"/", handler), Services())


def test_raises_for_route_mismatch_2():
    def handler(a: FromRoute[str]):
        ...

    with raises(RouteBinderMismatch):
        get_binders(Route(b"/:b", handler), Services())


def test_raises_for_unsupported_union():
    def handler(a: FromRoute[Union[str, int]]):
        ...

    with raises(NormalizationError):
        get_binders(Route(b"/:b", handler), Services())


def test_request_binding():
    def handler(request):
        ...

    binders = get_binders(Route(b"/", handler), Services())

    assert isinstance(binders[0], RequestBinder)


def test_services_binding():
    app_services = Services()

    def handler(services):
        assert services is app_services

    binders = get_binders(Route(b"/", handler), app_services)

    assert isinstance(binders[0], ExactBinder)


@pytest.mark.asyncio
async def test_services_from_normalization():
    app_services = Services()

    def handler(services):
        assert services is app_services
        return None

    method = normalize_handler(Route(b"/", handler), app_services)
    await method(None)


def test_raises_for_unsupported_signature():
    app_services = Services()

    def handler(services, *args):
        assert services is app_services
        return services

    def handler2(services, **kwargs):
        assert services is app_services
        return services

    def handler3(services, *, key_only):
        assert services is app_services
        return services

    with pytest.raises(UnsupportedSignatureError):
        normalize_handler(Route(b"/", handler), Services())

    with pytest.raises(UnsupportedSignatureError):
        normalize_handler(Route(b"/", handler2), Services())

    with pytest.raises(UnsupportedSignatureError):
        normalize_handler(Route(b"/", handler3), Services())


async def fake_handler(_):
    return "fake-handler-result"


@pytest.mark.asyncio
async def test_middleware_normalization():
    services = {"context": object()}
    fake_request = Request("GET", b"/", None)

    async def middleware(request, handler, context):
        assert request is fake_request
        assert handler is fake_handler
        assert context is services.get("context")
        return await handler(request)

    normalized = normalize_middleware(middleware, services)

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


@pytest.mark.asyncio
async def test_middleware_normalization_2():
    services = {"context": object(), "foo": object()}
    fake_request = Request("GET", b"/", None)

    async def middleware(context, foo):
        pass

    normalized = normalize_middleware(middleware, services)  # type: ignore

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


@pytest.mark.asyncio
async def test_middleware_normalization_raises_for_sync_function():
    def faulty_middleware(request, handler):
        pass

    with pytest.raises(ValueError):
        normalize_middleware(faulty_middleware, Services())  # type: ignore


@pytest.mark.asyncio
async def test_middleware_query_normalization():
    services = {"context": object()}
    fake_request = Request("GET", b"/?example=Lorem", None)

    async def middleware(request, handler, example: FromQuery[str]):
        assert request is fake_request
        assert handler is fake_handler
        assert example.value == "Lorem"
        return await handler(request)

    normalized = normalize_middleware(middleware, services)

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


@pytest.mark.asyncio
async def test_middleware_header_normalization():
    services = {"context": object()}
    fake_request = Request("GET", b"/", [(b"example", b"Lorem")])

    async def middleware(request, handler, example: FromHeader[str]):
        assert request is fake_request
        assert handler is fake_handler
        assert example.value == "Lorem"
        return await handler(request)

    normalized = normalize_middleware(middleware, services)

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


@pytest.mark.asyncio
async def test_middleware_normalization_no_parameters():
    services = {"context": object()}
    fake_request = Request("GET", b"/", [(b"example", b"Lorem")])

    called = False

    async def middleware():
        nonlocal called
        called = True

    normalized = normalize_middleware(middleware, services)  # type: ignore

    # NB: middlewares base signature is (request, handler)
    # since our middleware above does not handle the next request handler,
    # it is called by the normalized method
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert called
    assert result == "fake-handler-result"


def test_middleware_not_normalized_if_signature_matches_expected_signature():
    async def middleware(request, handler):
        return await handler(request)

    normalized = normalize_middleware(middleware, Services())
    assert normalized is middleware


def test_get_raw_bound_value_type_fallsback_to_str():
    class Foo:
        ...

    assert _get_raw_bound_value_type(Foo) is str  # type: ignore


def test_normalization_with_service_json_route_param():
    def handler(
        foo_id: str,
        data: FromJSON[Cat],
        some_service: SomeService,
    ):
        ...

    container = Container()
    container.add_transient(SomeService)

    binders = get_binders(Route("/{foo_id}", handler), container.build_provider())
    assert len(binders) == 3

    assert isinstance(binders[0], RouteBinder)
    assert binders[0].expected_type is str

    assert isinstance(binders[1], JSONBinder)
    assert binders[1].expected_type is Cat

    assert isinstance(binders[2], ServiceBinder)
    assert binders[2].expected_type is SomeService


def test_camel_case_route_parameter():
    def handler(statusKey: FromRoute[str]):
        ...

    binders = get_binders(Route(b"/:statusKey", handler), Container().build_provider())

    assert isinstance(binders[0], RouteBinder)
    assert binders[0].parameter_name == "statusKey"


def test_pascal_case_route_parameter():
    def handler(StatusKey: FromRoute[str]):
        ...

    binders = get_binders(Route(b"/:StatusKey", handler), Container().build_provider())

    assert isinstance(binders[0], RouteBinder)
    assert binders[0].parameter_name == "StatusKey"


def test_normalization_with_parameter_alias():
    class CustomBinder(Binder):
        name_alias = "sunn_o"

        def __init__(self):
            super().__init__(str)

        async def get_value(self, request: Request) -> Optional[str]:
            return "sunn_o"

    def handler(sunn_o):
        ...

    container = Container()

    binders = get_binders(Route("/", handler), container.build_provider())
    assert len(binders) == 1

    assert isinstance(binders[0], CustomBinder)



from .session import CircularRedirectError as CircularRedirectError
from .session import ClientSession as ClientSession
from .session import ConnectionTimeout as ConnectionTimeout
from .session import MaximumRedirectsExceededError as MaximumRedirectsExceededError
from .session import MissingLocationForRedirect as MissingLocationForRedirect
from .session import RequestTimeout as RequestTimeout



import httptools
from httptools.parser import errors


cdef class InvalidURL(Exception):
    def __init__(self, str message):
        super().__init__(message)


def valid_schema(schema):
    if schema and schema != b'https' and schema != b'http':
        raise InvalidURL(f'Expected http or https schema; got instead {schema.decode()}')


cdef class URL:

    def __init__(self, bytes value):
        cdef bytes schema
        cdef object port

        try:
            # if the value starts with a dot, prepend a slash;
            # urllib.parse urlparse handles those, while httptools raises
            # an exception
            if value and value[0] == 46:
                value = b"/" + value
            parsed = httptools.parse_url(value)
        except errors.HttpParserInvalidURLError:
            raise InvalidURL(f'The value cannot be parsed as URL ({value.decode()})')
        schema = parsed.schema
        valid_schema(schema)

        self.value = value or b''
        self.schema = schema
        self.host = parsed.host
        self.port = parsed.port or 0
        self.path = parsed.path
        self.query = parsed.query
        self.fragment = parsed.fragment
        self.is_absolute = parsed.schema is not None

    def __repr__(self):
        return f'<URL {self.value}>'

    def __str__(self):
        return self.value.decode()

    cpdef URL join(self, URL other):
        if other.is_absolute:
            raise ValueError(f'Cannot concatenate to an absolute URL ({self.value} + {other.value})')
        if self.query or self.fragment:
            raise ValueError('Cannot concatenate a URL with query or fragment to another URL portion')
        first_part = self.value
        other_part = other.value
        if first_part and other_part and first_part[-1] == 47 and other_part[0] == 47:
            return URL(first_part[:-1] + other_part)
        return URL(first_part + other_part)

    cpdef URL base_url(self):
        if not self.is_absolute:
            raise ValueError('This URL is relative. Cannot extract a base URL (without path).')
        cdef bytes base_url

        base_url = self.schema + b'://' + self.host

        if self.port != 0:
            if (self.schema == b'http' and self.port != 80) or (self.schema == b'https' and self.port != 443):
                base_url = base_url + b':' + str(self.port).encode()

        return URL(base_url)

    cpdef URL with_host(self, bytes host):
        if not self.is_absolute:
            raise TypeError("Cannot generate a URL from a partial URL")
        query = b"?" + self.query if self.query else b""
        fragment = b"#" + self.fragment if self.fragment else b""
        return URL(self.schema + b"://" + host + self.path + query + fragment)

    cpdef URL with_scheme(self, bytes schema):
        valid_schema(schema)

        if not self.is_absolute:
            raise TypeError("Cannot generate a URL from a partial URL")

        return URL(schema + self.value[len(self.schema):])

    def __add__(self, other):
        if isinstance(other, bytes):
            return self.join(URL(other))

        if isinstance(other, URL):
            return self.join(other)
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, URL):
            return self.value == other.value
        return NotImplemented


cpdef URL build_absolute_url(
    bytes scheme,
    bytes host,
    bytes base_path,
    bytes path
):
    valid_schema(scheme)
    return URL(
        scheme
        + b"://"
        + host
        + (b"/" if base_path else b"") + base_path.lstrip(b"/").rstrip(b"/")
        + (b"/" if path else b"") + path.lstrip(b"/")
    )



import os


def truthy(value: str, default: bool = False) -> bool:
    if not value:
        return default
    return value.upper() in {"1", "TRUE"}


class EnvironmentSettings:
    show_error_details: bool
    mount_auto_events: bool

    def __init__(self) -> None:
        self.show_error_details = truthy(os.environ.get("APP_SHOW_ERROR_DETAILS", ""))
        self.mount_auto_events = truthy(os.environ.get("APP_MOUNT_AUTO_EVENTS", ""))



from email.utils import formatdate

from blacksheep.contents import Content
from blacksheep.messages import Request, Response


def get_response_for_static_content(
    request: Request,
    content_type: bytes,
    contents: bytes,
    last_modified_time: float,
    cache_time: int = 10800,
) -> Response:
    """
    Returns a response object to serve static content.
    """
    current_etag = str(last_modified_time).encode()
    previous_etag = request.if_none_match

    headers = [
        (b"Last-Modified", formatdate(last_modified_time, usegmt=True).encode()),
        (b"ETag", current_etag),
    ]

    if cache_time > -1:
        headers.append((b"Cache-Control", b"max-age=" + str(cache_time).encode()))

    if previous_etag and current_etag == previous_etag:
        # handle HTTP 304 Not Modified
        return Response(304, headers, None)

    if request.method == "HEAD":
        headers.append((b"Content-Type", content_type))
        headers.append((b"Content-Length", str(len(contents)).encode()))
        return Response(200, headers, None)

    return Response(200, headers, Content(content_type, contents))



from enum import Enum
from functools import wraps
from typing import Any, AnyStr, Callable, List, MutableMapping, Optional

from blacksheep.messages import Request
from blacksheep.plugins import json
from blacksheep.server.asgi import get_full_path


class WebSocketState(Enum):
    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class MessageMode(Enum):
    TEXT = "text"
    BYTES = "bytes"


class WebSocketError(Exception):
    """Base class for all web sockets errors."""


class InvalidWebSocketStateError(WebSocketError):
    def __init__(
        self,
        *,
        party: str = "client",
        current_state: WebSocketState,
        expected_state: WebSocketState,
    ):
        super().__init__(party, current_state, expected_state)
        self.party = party
        self.current_state = current_state
        self.expected_state = expected_state

    def __str__(self):
        return (
            f"Invalid {self.party} state of the WebSocket connection. "
            f"Expected state: {self.expected_state}. "
            f"Current state: {self.current_state}."
        )


class WebSocketDisconnectError(WebSocketError):
    def __init__(self, code: int = 1000):
        super().__init__(code)
        self.code = code

    def __str__(self):
        return f"The client closed the connection. WebSocket close code: {self.code}."


class WebSocket(Request):
    def __init__(
        self, scope: MutableMapping[str, Any], receive: Callable, send: Callable
    ):
        assert scope["type"] == "websocket"
        super().__init__("GET", get_full_path(scope), list(scope["headers"]))

        self.scope = scope  # type: ignore
        self._receive = self._wrap_receive(receive)
        self._send = send
        self.route_values = {}

        self.client_state = WebSocketState.CONNECTING
        self.application_state = WebSocketState.CONNECTING

    def __repr__(self):
        return f"<WebSocket {self.url.value.decode()}>"

    async def _connect(self) -> None:
        if self.client_state != WebSocketState.CONNECTING:
            raise InvalidWebSocketStateError(
                current_state=self.client_state,
                expected_state=WebSocketState.CONNECTING,
            )

        message = await self._receive()
        assert message["type"] == "websocket.connect"

        self.client_state = WebSocketState.CONNECTED

    async def accept(
        self, headers: Optional[List] = None, subprotocol: str = None
    ) -> None:
        headers = headers or []

        await self._connect()
        self.application_state = WebSocketState.CONNECTED

        message = {
            "type": "websocket.accept",
            "headers": headers,
            "subprotocol": subprotocol,
        }

        await self._send(message)

    async def receive(self) -> MutableMapping[str, AnyStr]:
        if self.application_state != WebSocketState.CONNECTED:
            raise InvalidWebSocketStateError(
                party="application",
                current_state=self.application_state,
                expected_state=WebSocketState.CONNECTED,
            )

        message = await self._receive()
        assert message["type"] == "websocket.receive"

        return message

    async def receive_text(self) -> str:
        message = await self.receive()
        return message["text"]

    async def receive_bytes(self) -> bytes:
        message = await self.receive()
        return message["bytes"]

    async def receive_json(
        self, mode: MessageMode = MessageMode.TEXT
    ) -> MutableMapping[str, Any]:
        message = await self.receive()

        if mode == MessageMode.TEXT:
            return json.loads(message["text"])

        if mode == MessageMode.BYTES:
            return json.loads(message["bytes"].decode())

    async def _send_message(self, message: MutableMapping[str, AnyStr]) -> None:
        if self.client_state != WebSocketState.CONNECTED:
            raise InvalidWebSocketStateError(
                current_state=self.client_state,
                expected_state=WebSocketState.CONNECTED,
            )
        await self._send(message)

    async def send_text(self, data: str) -> None:
        await self._send_message({"type": "websocket.send", "text": data})

    async def send_bytes(self, data: bytes) -> None:
        await self._send_message({"type": "websocket.send", "bytes": data})

    async def send_json(
        self, data: MutableMapping[Any, Any], mode: MessageMode = MessageMode.TEXT
    ):
        text = json.dumps(data)

        if mode == MessageMode.TEXT:
            return await self.send_text(text)

        if mode == MessageMode.BYTES:
            return await self.send_bytes(text.encode())

    def _wrap_receive(self, _receive: Callable):
        @wraps(_receive)
        async def disconnect():
            message = await _receive()

            if message["type"] == "websocket.disconnect":
                self.application_state = self.client_state = WebSocketState.DISCONNECTED
                raise WebSocketDisconnectError(message["code"])

            return message

        return disconnect

    async def close(self, code: int = 1000, reason: Optional[str] = None) -> None:
        await self._send({"type": "websocket.close", "code": code, "reason": reason})



import pytest

from blacksheep.url import URL, InvalidURL


def test_empty_url():
    with pytest.raises(InvalidURL):
        URL(b"")


def test_absolute_url():
    url = URL(b"https://robertoprevato.github.io?foo=power&hello=world")

    assert url.path is None
    assert url.schema == b"https"
    assert url.host == b"robertoprevato.github.io"
    assert url.port == 0
    assert url.query == b"foo=power&hello=world"
    assert url.fragment is None
    assert url.is_absolute is True


def test_relative_url():
    url = URL(b"/api/cat/001?foo=power&hello=world")

    assert url.path == b"/api/cat/001"
    assert url.schema is None
    assert url.host is None
    assert url.port == 0
    assert url.query == b"foo=power&hello=world"
    assert url.fragment is None
    assert url.is_absolute is False


def test_equality():
    assert URL(b"/") == URL(b"/")
    assert URL(b"/api/cats") == URL(b"/api/cats")
    assert URL(b"/api/cats") != URL(b"/api/cat/001")


def test_concatenation():
    assert URL(b"/") + URL(b"/") == URL(b"/")
    assert URL(b"https://world-cats.eu") + URL(b"/api/cats") == URL(
        b"https://world-cats.eu/api/cats"
    )
    assert URL(b"https://world-cats.eu/") + URL(b"/api/cats") == URL(
        b"https://world-cats.eu/api/cats"
    )


def test_cannot_concatenate_one_url_first_with_query_or_path():
    with pytest.raises(ValueError):
        URL(b"https://world-cats.eu?hello=world") + URL(b"/api/cats")

    with pytest.raises(ValueError):
        URL(b"http://world-cats.eu/#/about") + URL(b"/api/cats")


def test_cannot_concatenate_absolute_urls():
    with pytest.raises(ValueError):
        URL(b"https://world-cats.eu") + URL(b"https://hello-world")

    with pytest.raises(ValueError):
        URL(b"http://world-cats.eu") + URL(b"http://hello-world")


@pytest.mark.parametrize(
    "value,expected_base_url",
    [
        [
            b"https://robertoprevato.github.io/api/v1/cats",
            b"https://robertoprevato.github.io",
        ],
        [
            b"https://robertoprevato.github.io:44555/api/v1/cats",
            b"https://robertoprevato.github.io:44555",
        ],
        [
            b"https://robertoprevato.github.io/api/v1/cats?lorem=ipsum",
            b"https://robertoprevato.github.io",
        ],
        [
            b"https://robertoprevato.github.io?lorem=ipsum",
            b"https://robertoprevato.github.io",
        ],
        [
            b"http://robertoprevato.github.io/api/v1/cats",
            b"http://robertoprevato.github.io",
        ],
        [
            b"http://robertoprevato.github.io:44555/api/v1/cats",
            b"http://robertoprevato.github.io:44555",
        ],
        [
            b"http://robertoprevato.github.io/api/v1/cats?lorem=ipsum",
            b"http://robertoprevato.github.io",
        ],
        [
            b"http://robertoprevato.github.io?lorem=ipsum",
            b"http://robertoprevato.github.io",
        ],
    ],
)
def test_base_url(value, expected_base_url):
    url = URL(value)
    base_url = url.base_url()
    assert expected_base_url == base_url.value


def test_raises_for_invalid_scheme():
    with pytest.raises(InvalidURL):
        URL(b"file://D:/a/b/c")



import html
import os
from pathlib import Path
from typing import Awaitable, Callable, Dict, Iterable, Optional, Sequence, Set
from urllib.parse import unquote

from blacksheep import HTMLContent, Request, Response
from blacksheep.common.files.asyncfs import FilesHandler
from blacksheep.common.files.pathsutils import get_file_extension_from_name
from blacksheep.exceptions import NotFound
from blacksheep.server.authorization import allow_anonymous
from blacksheep.server.resources import get_resource_file_content
from blacksheep.server.routing import Route, Router
from blacksheep.utils import join_fragments

from . import get_default_extensions, get_response_for_file, validate_source_path


def get_files_to_serve(
    source_folder: Path, extensions: Set[str], root_folder: Optional[Path] = None
) -> Iterable[Dict[str, str]]:
    assert source_folder.exists(), "The source folder path must exist"
    assert source_folder.is_dir(), "The source folder path must be a directory"

    if not root_folder:
        root_folder = source_folder

    names = os.listdir(source_folder)
    names.sort()
    dirs, nondirs = [], []

    for name in names:
        full_path = source_folder / name
        if os.path.isdir(full_path):
            dirs.append(full_path)
        else:
            nondirs.append(full_path)

    items = dirs + nondirs
    items = (item for item in items if not os.path.islink(item))

    yield {
        "rel_path": "../",
        "full_path": str(source_folder),
        "is_dir": True,
    }

    for item in items:
        item_path = str(item)

        if item.is_dir():
            yield {
                "rel_path": item_path[len(str(root_folder)) + 1 :],
                "full_path": item_path,
                "is_dir": True,
            }
        else:
            extension = get_file_extension_from_name(item_path)

            if extension in extensions:
                yield {
                    "rel_path": item_path[len(str(root_folder)) + 1 :],
                    "full_path": item_path,
                    "is_dir": False,
                }


def get_files_list_html_response(
    template: str,
    parent_folder_path: str,
    contents: Sequence[Dict[str, str]],
    root_path: str,
) -> Response:
    info_lines = []
    for item in contents:
        rel_path = item.get("rel_path")
        assert rel_path is not None
        full_rel_path = html.escape(
            join_fragments(root_path, parent_folder_path, rel_path)
        )
        info_lines.append(f'<li><a href="{full_rel_path}">{rel_path}</a></li>')
    info = "".join(info_lines)
    p = []
    whole_p = [root_path]
    for fragment in parent_folder_path.split("/"):
        if fragment:
            whole_p.append(html.escape(fragment))
            fragment_path = "/".join(whole_p)
            p.append(f'<a href="{fragment_path}">{html.escape(fragment)}</a>')

    # TODO: use chunked encoding here, yielding HTML fragments
    return Response(
        200,
        content=HTMLContent(template.format_map({"path": "/".join(p), "info": info})),
    )


def get_response_for_resource_path(
    request: Request,
    tail: str,
    files_list_html: str,
    source_folder_name: str,
    files_handler: FilesHandler,
    source_folder_full_path: str,
    discovery: bool,
    cache_time: int,
    extensions: Set[str],
    root_path: str,
    index_document: Optional[str],
) -> Response:
    resource_path = os.path.join(source_folder_name, tail)

    if "../" in tail:
        # verify that a relative path doesn't go outside of the
        # static root folder
        abs_path = os.path.abspath(resource_path)
        if not str(abs_path).lower().startswith(source_folder_full_path.lower()):
            # outside of the static folder!
            raise NotFound()

    if not os.path.exists(resource_path) or os.path.islink(resource_path):
        raise NotFound()

    if os.path.isdir(resource_path):
        # Request for a path that matches a folder: e.g. /foo/
        if discovery:
            return get_files_list_html_response(
                files_list_html,
                tail.rstrip("/"),
                list(get_files_to_serve(Path(resource_path.rstrip("/")), extensions)),
                root_path,
            )
        else:
            if index_document is not None:
                # try returning the default index document
                return get_response_for_resource_path(
                    request,
                    index_document,
                    files_list_html,
                    source_folder_name,
                    files_handler,
                    source_folder_full_path,
                    discovery,
                    cache_time,
                    extensions,
                    root_path,
                    None,
                )
            raise NotFound()

    file_extension = get_file_extension_from_name(resource_path)

    if file_extension not in extensions:
        raise NotFound()

    return get_response_for_file(files_handler, request, resource_path, cache_time)


def get_files_route_handler(
    files_handler: FilesHandler,
    source_folder_name: str,
    discovery: bool,
    cache_time: int,
    extensions: Set[str],
    root_path: str,
    index_document: Optional[str],
    fallback_document: Optional[str],
) -> Callable[[Request], Awaitable[Response]]:
    files_list_html = get_resource_file_content("fileslist.html")
    source_folder_full_path = os.path.abspath(str(source_folder_name))

    async def static_files_handler(request: Request) -> Response:
        assert request.route_values is not None, "Expects a route pattern with star *"
        tail = unquote(request.route_values.get("tail", "")).lstrip("/")

        try:
            return get_response_for_resource_path(
                request,
                tail,
                files_list_html,
                source_folder_name,
                files_handler,
                source_folder_full_path,
                discovery,
                cache_time,
                extensions,
                root_path,
                index_document,
            )
        except NotFound:
            if fallback_document is None:
                raise

            return get_response_for_resource_path(
                request,
                fallback_document,
                files_list_html,
                source_folder_name,
                files_handler,
                source_folder_full_path,
                discovery,
                cache_time,
                extensions,
                root_path,
                None,
            )

    return static_files_handler


def get_static_files_route(path_prefix: str) -> bytes:
    if not path_prefix:
        return b"*"
    if path_prefix[0] != "/":
        path_prefix = "/" + path_prefix
    if path_prefix[-1] != "/":
        path_prefix = path_prefix + "/"
    return path_prefix.encode() + b"*"


def serve_files_dynamic(
    router: Router,
    files_handler: FilesHandler,
    source_folder: str,
    *,
    discovery: bool,
    cache_time: int,
    extensions: Optional[Set[str]],
    root_path: str,
    index_document: Optional[str],
    fallback_document: Optional[str],
    anonymous_access: bool = True,
) -> None:
    """
    Configures a route to serve files dynamically, using the given files handler and
    options.
    """
    validate_source_path(source_folder)

    if not extensions:
        extensions = get_default_extensions()

    handler = get_files_route_handler(
        files_handler,
        str(source_folder),
        bool(discovery),
        int(cache_time),
        set(extensions),
        root_path,
        index_document,
        fallback_document,
    )

    if anonymous_access:
        handler = allow_anonymous()(handler)

    route = Route(
        get_static_files_route(root_path),
        handler,
    )
    router.add_route("GET", route)
    router.add_route("HEAD", route)



from typing import Optional

import pytest

from blacksheep import Content
from blacksheep.contents import JSONContent
from blacksheep.server.application import Application
from blacksheep.server.bindings import FromHeader
from blacksheep.server.controllers import Controller, RoutesRegistry
from blacksheep.server.responses import Response
from blacksheep.testing import AbstractTestSimulator, TestClient
from blacksheep.testing.helpers import (
    CookiesType,
    HeadersType,
    QueryType,
    get_example_scope,
)


class CustomTestSimulator(AbstractTestSimulator):
    async def send_request(
        self,
        method: str,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ):
        if method == "GET":
            return {"custom": "true"}


async def _start_application(app):
    await app.start()


@pytest.fixture
def test_app():
    return Application(show_error_details=True)


@pytest.mark.asyncio
async def test_client_response(test_app):
    @test_app.route("/")
    async def home(request):
        return {"foo": "bar"}

    await _start_application(test_app)

    test_client = TestClient(test_app)
    response = await test_client.get("/")

    actual_json_body = await response.json()
    expected_body = {"foo": "bar"}

    assert isinstance(response, Response)
    assert actual_json_body == expected_body


@pytest.mark.asyncio
async def test_client_headers(test_app):
    class FromTestHeader(FromHeader[str]):
        name = "test_header"

    @test_app.route("/")
    async def home(request, test_header: FromTestHeader):
        return test_header.value

    await _start_application(test_app)

    test_client = TestClient(test_app)
    response = await test_client.get("/", headers={"test_header": "foo"})

    actual_header_value = await response.text()
    expected_header_value = "foo"

    assert actual_header_value == expected_header_value


@pytest.mark.asyncio
async def test_client_content(test_app):
    @test_app.route("/", methods=["POST"])
    async def home(request):
        json_data = await request.json()
        return json_data

    await _start_application(test_app)

    test_client = TestClient(test_app)
    response = await test_client.post("/", content=JSONContent({"foo": "bar"}))

    actual_json_response = await response.json()
    expected_json_response = {"foo": "bar"}

    assert actual_json_response == expected_json_response


@pytest.mark.parametrize(
    "input_query",
    [
        {"foo": "bar"},
        [(b"foo", b"bar")],
        "foo=bar",
        b"foo=bar",
    ],
)
@pytest.mark.asyncio
async def test_client_queries(test_app, input_query):
    @test_app.route("/")
    async def home(request):
        return request.query

    await _start_application(test_app)

    test_client = TestClient(test_app)
    response = await test_client.get("/", query=input_query)

    actual_response = await response.json()
    expected_response = {"foo": ["bar"]}

    assert actual_response == expected_response


@pytest.mark.parametrize(
    "input_cookies",
    [
        {"foo": "bar"},
        [(b"foo", b"bar")],
    ],
)
@pytest.mark.asyncio
async def test_client_cookies(test_app, input_cookies):
    @test_app.route("/")
    async def home(request):
        return request.cookies

    await _start_application(test_app)

    test_client = TestClient(test_app)
    response = await test_client.get("/", cookies=input_cookies)

    actual_response = await response.json()
    expected_response = {"foo": "bar"}

    assert actual_response == expected_response


@pytest.mark.asyncio
async def test_client_content_raise_error_if_incorrect_type(test_app):
    with pytest.raises(ValueError):
        await _start_application(test_app)

        test_client = TestClient(test_app)

        await test_client.post("/", content={"foo": "bar"})


@pytest.mark.asyncio
async def test_client_application_not_started_error(test_app):
    with pytest.raises(AssertionError):
        TestClient(test_app)


@pytest.mark.asyncio
async def test_custom_test_simulator(test_app):
    test_client = TestClient(test_app, test_simulator=CustomTestSimulator())

    actual_response = await test_client.get("/")
    expected_response = {"custom": "true"}

    assert actual_response == expected_response


@pytest.mark.parametrize(
    "method, expected_method",
    [
        ("GET", "GET"),
        ("POST", "POST"),
        ("PATCH", "PATCH"),
        ("PUT", "PUT"),
        ("DELETE", "DELETE"),
        ("OPTIONS", "OPTIONS"),
        ("TRACE", "TRACE"),
        ("HEAD", "HEAD"),
    ],
)
@pytest.mark.asyncio
async def test_client_methods(test_app, method, expected_method):
    @test_app.route("/", methods=[method])
    async def home(request):
        return request.method

    await _start_application(test_app)

    client = TestClient(test_app)
    response = await getattr(client, method.lower())("/")

    actual_method = await response.text()

    assert actual_method == expected_method


def test_get_example_scope_raise_error_if_query_provided():
    with pytest.raises(ValueError):
        get_example_scope("GET", "/test?")


@pytest.mark.asyncio
async def test_app_controller_handle_correct_method(test_app):
    test_app.controllers_router = RoutesRegistry()
    get = test_app.controllers_router.get

    class Home(Controller):
        @get("/")
        async def index(self, request):
            assert isinstance(self, Home)
            return request.method

    await _start_application(test_app)

    client = TestClient(test_app)
    response = await client.get("/")

    expected_method = "GET"
    actual_method = await response.text()

    assert actual_method == expected_method


@pytest.mark.asyncio
async def test_app_controller_get_correct_post_body(test_app):
    test_app.controllers_router = RoutesRegistry()
    post = test_app.controllers_router.post

    class Home(Controller):
        @post("/")
        async def index(self, request):
            assert isinstance(self, Home)
            return await request.json()

    await _start_application(test_app)

    client = TestClient(test_app)
    response = await client.post("/", content=JSONContent({"foo": "bar"}))

    expected_json_body = {"foo": "bar"}
    actual_json_body = await response.json()

    assert actual_json_body == expected_json_body


@pytest.mark.asyncio
async def test_app_controller_get_correct_query_parameters(test_app):
    test_app.controllers_router = RoutesRegistry()
    get = test_app.controllers_router.get

    class Home(Controller):
        @get("/")
        async def index(self, request):
            assert isinstance(self, Home)
            return request.query

    await _start_application(test_app)

    client = TestClient(test_app)
    response = await client.get("/", query={"foo": "bar"})

    expected_query = {"foo": ["bar"]}
    actual_query = await response.json()

    assert actual_query == expected_query



from typing import Dict, Optional, Tuple

from guardpost.asynchronous.authentication import (
    AuthenticationHandler,
    AuthenticationStrategy,
)
from guardpost.authorization import AuthorizationError

from blacksheep import Response, TextContent

__all__ = (
    "AuthenticationStrategy",
    "AuthenticationHandler",
    "AuthenticateChallenge",
    "get_authentication_middleware",
    "handle_authentication_challenge",
)


def get_authentication_middleware(strategy: AuthenticationStrategy):
    async def authentication_middleware(request, handler):
        await strategy.authenticate(request, getattr(handler, "auth_schemes", None))
        return await handler(request)

    return authentication_middleware


class AuthenticateChallenge(AuthorizationError):
    header_name = b"WWW-Authenticate"

    def __init__(
        self, scheme: str, realm: Optional[str], parameters: Optional[Dict[str, str]]
    ):
        self.scheme = scheme
        self.realm = realm
        self.parameters = parameters

    def _get_header_value(self) -> bytes:
        if not self.realm and not self.parameters:
            return self.scheme.encode()

        parts = bytearray(self.scheme.encode())
        if self.realm:
            parts.extend(f' realm="{self.realm}"'.encode())

        if self.parameters:
            parts.extend(b", ")
            parts.extend(
                b", ".join(
                    [
                        f'{key}="{value}"'.encode()
                        for key, value in self.parameters.items()
                    ]
                )
            )
        return bytes(parts)

    def get_header(self) -> Tuple[bytes, bytes]:
        return self.header_name, self._get_header_value()


async def handle_authentication_challenge(
    app, request, exception: AuthenticateChallenge
):
    return Response(401, [exception.get_header()], content=TextContent("Unauthorized"))



import os

from blacksheep.server.env import EnvironmentSettings


def test_env_settings():
    os.environ["APP_SHOW_ERROR_DETAILS"] = "1"
    env = EnvironmentSettings()

    assert env.show_error_details is True

    os.environ["APP_SHOW_ERROR_DETAILS"] = "0"
    env = EnvironmentSettings()

    assert env.show_error_details is False

    os.environ["APP_SHOW_ERROR_DETAILS"] = "true"
    env = EnvironmentSettings()

    assert env.show_error_details is True



import pytest

from blacksheep.server.bindings import FromHeader
from blacksheep.server.websocket import (
    InvalidWebSocketStateError,
    MessageMode,
    WebSocket,
    WebSocketDisconnectError,
    WebSocketState,
)
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


@pytest.fixture
def example_scope():
    return {
        "type": "websocket",
        "path": "/ws",
        "query_string": "",
        "headers": [(b"upgrade", b"websocket")],
    }


def test_websocket_repr(example_scope):
    ws = WebSocket(example_scope, MockReceive([]), MockSend())

    assert str(ws) == "<WebSocket /ws>"


@pytest.mark.asyncio
async def test_connect_raises_if_not_connecting(example_scope):
    ws = WebSocket(
        example_scope, MockReceive([{"type": "websocket.connect"}]), MockSend()
    )

    ws.client_state = WebSocketState.CONNECTED

    with pytest.raises(InvalidWebSocketStateError) as error:
        await ws.accept()

    assert error.value.current_state == WebSocketState.CONNECTED
    assert error.value.expected_state == WebSocketState.CONNECTING

    assert str(error.value) == (
        f"Invalid {error.value.party} state of the WebSocket connection. "
        f"Expected state: {error.value.expected_state}. "
        f"Current state: {error.value.current_state}."
    )


@pytest.mark.asyncio
async def test_websocket_accept(example_scope):
    """
    A websocket gets fully connected when the ASGI server sends a message of type
    'websocket.connect' and the server accepts the connection.
    """
    ws = WebSocket(
        example_scope, MockReceive([{"type": "websocket.connect"}]), MockSend()
    )

    await ws.accept()

    assert ws.client_state == WebSocketState.CONNECTED
    assert ws.application_state == WebSocketState.CONNECTED


@pytest.mark.asyncio
async def test_websocket_receive_text(example_scope):
    """
    A first message is received when the underlying ASGI server first sends a
    'websocket.connect' message, then a content message.
    """
    ws = WebSocket(
        example_scope,
        MockReceive(
            [
                {"type": "websocket.connect"},
                {"type": "websocket.receive", "text": "Lorem ipsum dolor sit amet"},
            ]
        ),
        MockSend(),
    )

    await ws.accept()

    message = await ws.receive_text()
    assert message == "Lorem ipsum dolor sit amet"


@pytest.mark.asyncio
async def test_websocket_receive_bytes(example_scope):
    """
    A first message is received when the underlying ASGI server first sends a
    'websocket.connect' message, then a content message.
    """
    ws = WebSocket(
        example_scope,
        MockReceive(
            [
                {"type": "websocket.connect"},
                {"type": "websocket.receive", "bytes": b"Lorem ipsum dolor sit amet"},
            ]
        ),
        MockSend(),
    )

    await ws.accept()

    message = await ws.receive_bytes()
    assert message == b"Lorem ipsum dolor sit amet"


@pytest.mark.asyncio
async def test_websocket_receive_json(example_scope):
    """
    A first message is received when the underlying ASGI server first sends a
    'websocket.connect' message, then a content message.
    """
    ws = WebSocket(
        example_scope,
        MockReceive(
            [
                {"type": "websocket.connect"},
                {"type": "websocket.receive", "text": '{"message": "Lorem ipsum"}'},
            ]
        ),
        MockSend(),
    )

    await ws.accept()

    message = await ws.receive_json()
    assert message == {"message": "Lorem ipsum"}


@pytest.mark.asyncio
async def test_websocket_receive_json_from_bytes(example_scope):
    """
    A first message is received when the underlying ASGI server first sends a
    'websocket.connect' message, then a content message.
    """
    ws = WebSocket(
        example_scope,
        MockReceive(
            [
                {"type": "websocket.connect"},
                {"type": "websocket.receive", "bytes": b'{"message": "Lorem ipsum"}'},
            ]
        ),
        MockSend(),
    )

    await ws.accept()

    message = await ws.receive_json(mode=MessageMode.BYTES)
    assert message == {"message": "Lorem ipsum"}


@pytest.mark.asyncio
async def test_websocket_send_text(example_scope):
    """
    A message is sent by the server to clients, by sending a message to the underlying
    ASGI server with type "websocket.send" and a "text" or "bytes" property.
    """
    mocked_send = MockSend()
    ws = WebSocket(
        example_scope,
        MockReceive([{"type": "websocket.connect"}]),
        mocked_send,
    )

    await ws.accept()

    await ws.send_text("Lorem ipsum dolor sit amet")

    assert len(mocked_send.messages) > 0
    message = mocked_send.messages[-1]

    assert message.get("text") == "Lorem ipsum dolor sit amet"
    assert message.get("type") == "websocket.send"


@pytest.mark.asyncio
async def test_websocket_send_bytes(example_scope):
    """
    A message is sent by the server to clients, by sending a message to the underlying
    ASGI server with type "websocket.send" and a "text" or "bytes" property.
    """
    mocked_send = MockSend()
    ws = WebSocket(
        example_scope,
        MockReceive([{"type": "websocket.connect"}]),
        mocked_send,
    )

    await ws.accept()

    await ws.send_bytes(b"Lorem ipsum dolor sit amet")

    assert len(mocked_send.messages) > 0
    message = mocked_send.messages[-1]

    assert message.get("bytes") == b"Lorem ipsum dolor sit amet"
    assert message.get("type") == "websocket.send"


@pytest.mark.asyncio
async def test_websocket_send_json(example_scope):
    """
    A message is sent by the server to clients, by sending a message to the underlying
    ASGI server with type "websocket.send" and a "text" or "bytes" property.
    """
    mocked_send = MockSend()
    ws = WebSocket(
        example_scope,
        MockReceive([{"type": "websocket.connect"}]),
        mocked_send,
    )

    await ws.accept()

    await ws.send_json({"message": "Lorem ipsum dolor sit amet"})

    assert len(mocked_send.messages) > 0
    message = mocked_send.messages[-1]

    assert message.get("text") == '{"message":"Lorem ipsum dolor sit amet"}'
    assert message.get("type") == "websocket.send"


@pytest.mark.asyncio
async def test_websocket_send_json_as_bytes(example_scope):
    """
    A message is sent by the server to clients, by sending a message to the underlying
    ASGI server with type "websocket.send" and a "text" or "bytes" property.
    """
    mocked_send = MockSend()
    ws = WebSocket(
        example_scope,
        MockReceive([{"type": "websocket.connect"}]),
        mocked_send,
    )

    await ws.accept()

    await ws.send_json({"message": "Lorem ipsum dolor sit amet"}, MessageMode.BYTES)

    assert len(mocked_send.messages) > 0
    message = mocked_send.messages[-1]

    assert message.get("bytes") == b'{"message":"Lorem ipsum dolor sit amet"}'
    assert message.get("type") == "websocket.send"


@pytest.mark.asyncio
async def test_connecting_websocket_raises_for_receive(example_scope):
    ws = WebSocket(example_scope, MockReceive(), MockSend())

    assert ws.client_state == WebSocketState.CONNECTING

    with pytest.raises(InvalidWebSocketStateError) as error:
        await ws.receive()

    assert error.value.current_state == WebSocketState.CONNECTING
    assert error.value.expected_state == WebSocketState.CONNECTED

    assert str(error.value) == (
        f"Invalid {error.value.party} state of the WebSocket connection. "
        f"Expected state: {error.value.expected_state}. "
        f"Current state: {error.value.current_state}."
    )


@pytest.mark.asyncio
async def test_connecting_websocket_raises_for_send(example_scope):
    ws = WebSocket(example_scope, MockReceive(), MockSend())

    assert ws.client_state == WebSocketState.CONNECTING

    with pytest.raises(InvalidWebSocketStateError) as error:
        await ws.send_text("Error")

    assert error.value.current_state == WebSocketState.CONNECTING
    assert error.value.expected_state == WebSocketState.CONNECTED

    assert str(error.value) == (
        f"Invalid {error.value.party} state of the WebSocket connection. "
        f"Expected state: {error.value.expected_state}. "
        f"Current state: {error.value.current_state}."
    )


@pytest.mark.asyncio
async def test_websocket_raises_for_receive_when_closed_by_client(example_scope):
    """
    If the underlying ASGI server sends a message of type "websocket.disconnect",
    it means that the client disconnected.
    """
    ws = WebSocket(
        example_scope,
        MockReceive(
            [
                {"type": "websocket.connect"},
                {"type": "websocket.disconnect", "code": 500},
            ]
        ),
        MockSend(),
    )

    await ws.accept()

    with pytest.raises(WebSocketDisconnectError) as error:
        await ws.receive()

    assert error.value.code == 500

    assert str(error.value) == (
        f"The client closed the connection. WebSocket close code: {error.value.code}."
    )


@pytest.mark.asyncio
async def test_application_handling_websocket_request_not_found(example_scope):
    """
    If a client tries to open a WebSocket connection on an endpoint that is not handled,
    the application returns an ASGI message to close the connection.
    """
    app = FakeApplication()
    mock_send = MockSend()
    mock_receive = MockReceive()

    await app(example_scope, mock_receive, mock_send)

    close_message = mock_send.messages[0]
    assert close_message == {"type": "websocket.close", "reason": None, "code": 1000}


@pytest.mark.asyncio
async def test_application_handling_proper_websocket_request():
    """
    If a client tries to open a WebSocket connection on an endpoint that is handled,
    the application websocket handler is called.
    """
    app = FakeApplication()
    mock_send = MockSend()
    mock_receive = MockReceive([{"type": "websocket.connect"}])

    @app.router.ws("/ws/{foo}")
    async def websocket_handler(websocket, foo):
        assert isinstance(websocket, WebSocket)
        assert websocket.application_state == WebSocketState.CONNECTING
        assert websocket.client_state == WebSocketState.CONNECTING

        assert foo == "001"

        await websocket.accept()

    await app.start()
    await app(
        {"type": "websocket", "path": "/ws/001", "query_string": "", "headers": []},
        mock_receive,
        mock_send,
    )


@pytest.mark.asyncio
async def test_application_handling_proper_websocket_request_with_query():
    app = FakeApplication()
    mock_send = MockSend()
    mock_receive = MockReceive([{"type": "websocket.connect"}])

    @app.router.ws("/ws/{foo}")
    async def websocket_handler(websocket, foo, from_query: int):
        assert isinstance(websocket, WebSocket)
        assert websocket.application_state == WebSocketState.CONNECTING
        assert websocket.client_state == WebSocketState.CONNECTING

        assert foo == "001"
        assert from_query == 200

        await websocket.accept()

    await app.start()
    await app(
        {
            "type": "websocket",
            "path": "/ws/001",
            "query_string": b"from_query=200",
            "headers": [],
        },
        mock_receive,
        mock_send,
    )


@pytest.mark.asyncio
async def test_application_handling_proper_websocket_request_header_binding(
    example_scope,
):
    app = FakeApplication()
    mock_send = MockSend()
    mock_receive = MockReceive([{"type": "websocket.connect"}])

    class UpgradeHeader(FromHeader[str]):
        name = "Upgrade"

    called = False

    @app.router.ws("/ws")
    async def websocket_handler(connect_header: UpgradeHeader):
        assert connect_header.value == "websocket"

        nonlocal called
        called = True

    await app.start()
    await app(example_scope, mock_receive, mock_send)
    assert called is True


@pytest.mark.asyncio
async def test_application_websocket_binding_by_type_annotation():
    """
    This test verifies that the WebSocketBinder can bind a WebSocket by type annotation.
    """
    app = FakeApplication()
    mock_send = MockSend()
    mock_receive = MockReceive([{"type": "websocket.connect"}])

    @app.router.ws("/ws")
    async def websocket_handler(my_ws: WebSocket):
        assert isinstance(my_ws, WebSocket)
        assert my_ws.application_state == WebSocketState.CONNECTING
        assert my_ws.client_state == WebSocketState.CONNECTING

        await my_ws.accept()

    await app.start()
    await app(
        {"type": "websocket", "path": "/ws", "query_string": "", "headers": []},
        mock_receive,
        mock_send,
    )



import weakref
from typing import TYPE_CHECKING, Optional, Sequence, Tuple

from itsdangerous import Serializer
from itsdangerous.exc import BadSignature
from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.utils import pass_context

from blacksheep.baseapp import get_logger
from blacksheep.cookies import Cookie
from blacksheep.exceptions import Unauthorized
from blacksheep.messages import Request, Response
from blacksheep.server.application import Application
from blacksheep.server.dataprotection import generate_secret, get_serializer
from blacksheep.server.security import SecurityPolicyHandler


class AntiForgeryTokenError(Unauthorized):
    def __init__(self, message: str):
        super().__init__(message)


class InvalidAntiForgeryTokenError(AntiForgeryTokenError):
    def __init__(self, message: str = "Invalid anti-forgery token"):
        super().__init__(message)


class MissingRequestContextError(TypeError):
    def __init__(self) -> None:
        super().__init__(
            "The request context is missing from the render call. Pass the request "
            "object to the context of the template."
        )


class ClickJackingProtection(SecurityPolicyHandler):
    """
    Class used to protect responses from click-jacking. This class is applied by default
    when using Anti-Forgery validation.
    """

    def protect(self, response: Response) -> None:
        """
        Applies response headers, to configure Content-Security-Policy and
        X-Frame-Options.
        """
        # for modern browsers
        response.add_header(b"Content-Security-Policy", b"frame-ancestors: 'self';")

        # for old browsers
        response.add_header(b"X-Frame-Options", b"SAMEORIGIN")


class AntiForgeryHandler:
    def __init__(
        self,
        cookie_name: str = "aftoken",
        form_name: str = "__RequestVerificationToken",
        header_name: str = "RequestVerificationToken",
        secret_keys: Optional[Sequence[str]] = None,
        serializer: Optional[Serializer] = None,
        security_handler: Optional[SecurityPolicyHandler] = None,
    ) -> None:
        """
        Creates a new instance of AntiForgeryHandler, that validates incoming
        requests that should provide an AntiForgery token.

        Parameters
        ----------
        cookie_name : str, optional
            The name of the cookie used to restore user's identity, by default
            "aftoken"
        header_name : str, optional
            The name of the header that can be used to send the control value, by
            default "RequestVerificationToken"
        form_name : str, optional
            The name of the form input that can be used to send the control value, by
            default "__RequestVerificationToken"
        secret_key : str, optional
            If specified, the key used by a default serializer (when no serializer is
            specified), by default None
        serializer : Optional[Serializer], optional
            If specified, controls the serializer used to sign and verify the values
            of cookies used for identities, by default None
        security_handler : Optional[SecurityPolicyHandler], optional
            Object used to protect responses that contain anti-forgery tokens. By
            default documents are protected against click-jacking allowing iframes only
            from the same site. Provide a custom object to control security rules.
        """
        self.cookie_name = cookie_name
        self.header_name = header_name
        self.form_name = form_name
        self._serializer = serializer or get_serializer(secret_keys, "antiforgery")
        self.tokens = weakref.WeakKeyDictionary()
        self.reuse_tokens_among_requests = True
        self.security_policy = security_handler or ClickJackingProtection()
        self.logger = get_logger()

    VALIDATE_METHODS = set("PATCH POST PUT DELETE".split())

    @property
    def serializer(self) -> Serializer:
        return self._serializer

    async def __call__(self, request: Request, handler):
        if self._should_validate_route(request, handler):
            try:
                await self.validate_request(request)
            except AntiForgeryTokenError as token_error:
                return Response(401, [(b"Reason", str(token_error).encode())])

        response: Response = await handler(request)

        # are there tokens issued for the current request?
        # TODO: use walrus := when support for Python 3.7 is dropped
        tokens = self.tokens.get(request)

        if tokens:
            self.set_cookie(response, tokens[0])
            self.security_policy.protect(response)

        return response

    def _should_validate_route(self, request: Request, handler) -> bool:
        if _is_ignored_handler(handler):
            return False

        return self.should_validate_request(request)

    def should_validate_request(self, request: Request) -> bool:
        """
        Returns a value indicating whether the given request should be validated.
        Note that, even though according to the specification HTML form methods can
        only be GET or POST, this class by default validates requests also for other
        methods that can write / delete information: PATCH PUT DELETE.
        """
        return request.method in self.VALIDATE_METHODS

    async def validate_request(self, request: Request) -> None:
        token_in_cookie = self.read_cookie(request)
        token_in_second_parameter = await self.read_control_value(request)

        try:
            cookie_value = self.serializer.loads(token_in_cookie)
            control_value = self.serializer.loads(token_in_second_parameter)
        except BadSignature:
            raise InvalidAntiForgeryTokenError()

        if cookie_value != control_value[::-1]:
            self.logger.info("Invalid Anti-Forgery token: values do not match.")
            raise InvalidAntiForgeryTokenError()

    def set_cookie(self, response: Response, value: str, secure: bool = False) -> None:
        """
        Sets the cookie used to store the Anti-Forgery token. When validating requests,
        this token is read from the cookie and and matched with a second value that was
        transmitted in other ways (e.g. inside an input element).

        Parameters
        ----------
        data : Any
            Anything that can be serialized by an `itsdangerous.Serializer`, a
            dictionary in the most common scenario.
        response : Response
            The instance of blacksheep `Response` that will include the cookie for the
            client.
        secure : bool, optional
            Whether the set cookie should have secure flag, by default False
        """
        response.set_cookie(
            Cookie(
                self.cookie_name,
                value,
                domain=None,
                path="/",
                http_only=True,
                secure=secure,
            )
        )

    def read_cookie(self, request: Request) -> str:
        """
        Reads the value of the Anti-Forgery token stored in the cookie.
        This method raises an AntiForgeryTokenError if the cookie is missing.
        """
        cookie = request.get_cookie(self.cookie_name)

        if cookie is None:
            raise AntiForgeryTokenError("Missing anti-forgery token cookie")

        return cookie

    async def read_control_value(self, request: Request) -> str:
        """
        Reads the control-value token stored in a second location.
        By default this code tries to read the anti-forgery token from a header,
        """
        control_value = request.headers.get_first(self.header_name.encode())

        if control_value is not None:
            return control_value.decode()

        form_data = await request.form()

        if form_data is not None and self.form_name in form_data:
            value = form_data[self.form_name]
            if not isinstance(value, str):
                # the value can only be a list, this happens when more than one input
                # is configured (for example a mistake where {% af_input %} appears
                # more than once in the same form; use only the first value
                assert isinstance(value, list)
                return value[0]
            return value

        raise AntiForgeryTokenError("Missing anti-forgery token control value")

    def get_tokens(self, request: Request) -> Tuple[str, str]:
        """
        Gets a pair of anti-forgery tokens for the given request.
        If tokens were already generated for this request, it returns them, otherwise
        creates new tokens. By default, if tokens were already generated for the same
        client (i.e. if the web request contains a request cookie of a previous
        generation), it returns by default the same values to support the same page
        opened in multiple tabs of the browser, because generating a new cookie would
        invalidate the previous page.

        When Anti-Forgery tokens are generated for a web request,
        this class also configures a matching cookie in the generated response, to send
        one of the two values that will be used to validate subsequent web requests.
        """
        existing_cookie = request.cookies.get(self.cookie_name)

        if existing_cookie is not None and self.reuse_tokens_among_requests:
            # Do not generate new tokens for the same client. This is to support the
            # same web site being opened in multiple tabs of the same browser without
            # breaking any of those pages. This behavior can be disabled setting
            # reuse_tokens_among_requests to False.
            value_in_cookie = existing_cookie

            try:
                parsed_value = self.serializer.loads(value_in_cookie)
            except BadSignature:
                """Do nothing in this case, generate new tokens."""
            else:
                control_value = self._generate_control_token(parsed_value)

                # Note: in this case we don't set self.tokens[request] to not set a new
                # cookie on the Response object! We don't need it anyway.
                return (value_in_cookie, control_value)

        tokens = self.tokens.get(request)

        if tokens:
            return tokens
        else:
            tokens = self._generate_tokens()
            self.tokens[request] = tokens
            return tokens

    def _generate_control_token(self, cookie_value: str) -> str:
        control_value = self.serializer.dumps(cookie_value[::-1])
        assert isinstance(control_value, str)
        return control_value

    def _generate_tokens(self) -> Tuple[str, str]:
        """
        Generates two different tokens that can be matched later:
        * one is set as HTTPOnly cookie
        * one will be served in another way in HTML, for example as input element in
          forms
        """
        token = generate_secret(40)

        value_one = self.serializer.dumps(token)
        assert isinstance(value_one, str)

        value_two = self._generate_control_token(token)
        return (value_one, value_two)


class AntiForgeryBaseExtension(Extension):
    af_handler: Optional[AntiForgeryHandler] = None

    def __init__(self, environment):
        super().__init__(environment)

        if self.af_handler is None:
            raise TypeError("Define a subclass bound to an AntiForgeryHandler")

    @property
    def handler(self) -> AntiForgeryHandler:
        if self.af_handler is None:
            raise TypeError("Missing anti_forgery_handler")
        return self.af_handler

    def parse(self, parser):
        line_number = next(parser.stream).lineno
        return nodes.CallBlock(self.call_method("get_html"), [], [], "").set_lineno(
            line_number
        )

    def get_token(self, context) -> str:
        try:
            request = context["request"]
        except KeyError:
            raise MissingRequestContextError()
        assert isinstance(request, Request)

        tokens = self.handler.get_tokens(request)
        return tokens[1]


class AntiForgeryInputExtension(AntiForgeryBaseExtension):
    tags = {"csrf_input", "af_input"}

    @pass_context
    def get_html(self, context, caller):
        value = self.get_token(context)
        return (
            f'<input type="hidden" name="{self.handler.form_name}" '
            f'value="{value}" />'
        )


class AntiForgeryValueExtension(AntiForgeryBaseExtension):
    tags = {"csrf_token", "af_token"}

    @pass_context
    def get_html(self, context, caller):
        return self.get_token(context)


_IGNORED_HANDLERS = {}


def _is_ignored_handler(handler) -> bool:
    """
    Returns a value indicating whether the given request handler is explicitly ignored
    for anti-forgery validation (for example to support cases that use alternative
    ways to authenticate the request, not vulnerable to CRSF).
    """
    if (
        handler in _IGNORED_HANDLERS
        or getattr(handler, "root_fn", None) in _IGNORED_HANDLERS
    ):
        return True

    return False


def ignore_anti_forgery(value: bool = True):
    """Optionally excludes a request handler from anti forgery validation."""

    def decorator(fn):
        _IGNORED_HANDLERS[fn] = value
        return fn

    return decorator


def use_anti_forgery(
    app: Application, handler: Optional[AntiForgeryHandler] = None
) -> AntiForgeryHandler:
    """
    Configures Anti-Forgery validation on the given application, to protect against
    Cross-Site Request Forgery (XSRF/CSRF) attacks.

    If the application is configured to use Jinja2, this function applies extensions to
    the templating environment to support rendering an anti-forgery token into the HTML
    of a web document.

    Example to render a hidden input:

        <form action="/user" method="post">
            {% af_input %}
            <input type="text" name="username" />
            <input type="submit" value="Submit" />
        </form>

    Example to render only a value:

        <script>
            EXAMPLE = {"token": "{% af_token %}"}
        </script>

    When an anti-forgery token is rendered in a view, the HTTP Response object receives
    also a cookie with a control value.
    """
    if handler is None:
        handler = AntiForgeryHandler()

    env = getattr(app, "templates_environment")

    if env is None:  # pragma: no cover
        handler.logger.info(
            "Templating is not configured on the application, extensions to render "
            "anti-forgery tokens with Jinja2 won't be configured."
        )
    else:
        if TYPE_CHECKING:  # pragma: no cover
            from jinja2 import Environment

            assert isinstance(env, Environment)

        class BoundAntiForgeryInputExtension(AntiForgeryInputExtension):
            af_handler = handler

        class BoundAntiForgeryValueExtension(AntiForgeryValueExtension):
            af_handler = handler

        env.add_extension(BoundAntiForgeryInputExtension)
        env.add_extension(BoundAntiForgeryValueExtension)

    app.middlewares.append(handler)

    return handler



"""
This module provides classes to handle OpenID Connect authentication through integration
with OAuth applications, supporting Authorization Code Grant and Hybrid flows.
"""
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, AnyStr, Awaitable, Callable, Dict, Optional, Sequence
from urllib.parse import urlencode

from guardpost.authorization import UnauthorizedError
from guardpost.jwts import InvalidAccessToken, JWTValidator
from itsdangerous import Serializer
from itsdangerous.exc import BadSignature
from jwt import InvalidTokenError

from blacksheep.cookies import Cookie
from blacksheep.exceptions import BadRequest, Unauthorized
from blacksheep.messages import Request, Response, get_absolute_url_to_path
from blacksheep.server.application import Application, ApplicationEvent
from blacksheep.server.authentication.cookie import CookieAuthentication
from blacksheep.server.authorization import allow_anonymous
from blacksheep.server.dataprotection import generate_secret, get_serializer
from blacksheep.server.responses import accepted, redirect
from blacksheep.utils import ensure_str
from blacksheep.utils.aio import FailedRequestError, HTTPHandler


def get_logger() -> logging.Logger:
    logger = logging.getLogger("blacksheep.oidc")
    logger.setLevel(logging.INFO)
    return logger


logger = get_logger()


class OpenIDConfiguration:
    """
    Proxy class for a remote OpenID Connect well-known configuration.
    """

    def __init__(self, data) -> None:
        self._data = data

    @property
    def issuer(self) -> str:
        return self._data["issuer"]

    @property
    def jwks_uri(self) -> str:
        return self._data["jwks_uri"]

    @property
    def authorization_endpoint(self) -> str:
        return self._data["authorization_endpoint"]

    @property
    def token_endpoint(self) -> str:
        return self._data["token_endpoint"]

    @property
    def end_session_endpoint(self) -> Optional[str]:
        return self._data.get("end_session_endpoint")


class OpenIDConnectEvent(ApplicationEvent):
    pass


class OpenIDConnectEvents:
    on_id_token_validated: OpenIDConnectEvent
    on_tokens_received: OpenIDConnectEvent
    on_error: OpenIDConnectEvent

    def __init__(self, context) -> None:
        self.on_tokens_received = OpenIDConnectEvent(context)
        self.on_id_token_validated = OpenIDConnectEvent(context)
        self.on_error = OpenIDConnectEvent(context)


@dataclass
class OpenIDSettings:
    client_id: str
    authority: Optional[str] = None
    audience: Optional[str] = None
    client_secret: Optional[str] = None
    discovery_endpoint: Optional[str] = None
    entry_path: str = "/sign-in"
    logout_path: str = "/sign-out"
    post_logout_redirect_path: str = "/"
    callback_path: str = "/authorization-callback"
    response_type: str = "code"
    scope: str = "openid profile email"
    redirect_uri: Optional[str] = None
    scheme_name: str = "OpenIDConnect"
    error_redirect_path: Optional[str] = None
    end_session_endpoint: Optional[str] = None


class OpenIDSettingsError(TypeError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class MissingClientSecretSettingError(OpenIDSettingsError):
    def __init__(self) -> None:
        super().__init__(
            "Missing application client secret, to use the Authorization Code Grant "
            "flow, it is necessary to configure an application secret."
        )


def _get_desired_path(request: Request) -> str:
    base_path = request.base_path
    path = request.path
    query_part = (
        "" if not request.url.query else ("?" + request.url.query.decode("utf8"))
    )
    return f"{base_path}{path}{query_part}"


class ParametersBuilder:
    def __init__(self, settings: OpenIDSettings) -> None:
        self._settings = settings
        self._serializer = get_serializer(purpose=f"{settings.scheme_name}state")

    @property
    def scope(self) -> str:
        return self._settings.scope

    def get_state(self, request: Request) -> Dict[str, str]:
        desired_path = (
            "/"
            if request.path == self._settings.entry_path
            else _get_desired_path(request)
        )
        return {"orig_path": desired_path, "nonce": generate_secret(8)}

    def read_state(self, state: str) -> Dict[str, str]:
        try:
            return self._serializer.loads(state)
        except BadSignature as signature_error:
            logger.error(
                "Failed to parse the request state (%s), invalid signature.",
                str(signature_error),
            )
            raise BadRequest("Invalid state")

    def get_redirect_url(self, request: Request) -> str:
        return str(get_absolute_url_to_path(request, self._settings.callback_path))

    def build_signin_parameters(self, request: Request):
        if self._settings.client_secret:
            # authorization code grant
            response_type = "code"
        else:
            # hybrid flow, requires implicit flow for id_token to be enabled
            response_type = "id_token"

        state = self.get_state(request)
        parameters = {
            "response_type": response_type,
            "response_mode": "form_post",
            "scope": self.scope,
            "client_id": self._settings.client_id,
            "redirect_uri": self._settings.redirect_uri
            or self.get_redirect_url(request),
            "nonce": state.get("nonce") or generate_secret(8),
        }

        if self._settings.audience:
            # Note: Auth0 and Okta use `audience` parameter when
            # the scope includes custom scopes
            parameters["audience"] = self._settings.audience

        if state:
            parameters["state"] = self._serializer.dumps(state)  # type: ignore

        return parameters

    def _require_secret(self):
        if not self._settings.client_secret:
            raise MissingClientSecretSettingError()

    def build_code_grant_parameters(self, request: Request, code: str):
        self._require_secret()

        return {
            "grant_type": "authorization_code",
            "code": code,
            "scope": self.scope,
            "redirect_uri": self._settings.redirect_uri
            or self.get_redirect_url(request),
            "client_id": self._settings.client_id,
            "client_secret": self._settings.client_secret,
        }

    def build_refresh_token_parameters(self, refresh_token: str):
        self._require_secret()

        return {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self._settings.client_id,
            "client_secret": self._settings.client_secret,
        }


class OpenIDConnectError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class OpenIDConnectConfigurationError(OpenIDConnectError):
    """
    Exception thrown when there is an error in the programmer's
    defined configuration.
    """


class OpenIDConnectRequestError(OpenIDConnectError):
    def __init__(
        self,
        request_error: FailedRequestError,
        message: str = "Failed web request to the OpenIDConnect provider.",
    ) -> None:
        super().__init__(message + " " + str(request_error))
        self.request_error = request_error


class OpenIDConnectFailedExchangeError(OpenIDConnectRequestError):
    def __init__(
        self,
        request_error: FailedRequestError,
        message: str = (
            "Failed token exchange web request to the OpenIDConnect provider."
        ),
    ) -> None:
        super().__init__(request_error, message=message)


class TokenResponse:
    __slots__ = ("data",)

    def __init__(self, data) -> None:
        self.data = data

    def __repr__(self) -> str:
        return repr(self.data)

    def __str__(self) -> str:
        return str(self.data)

    @property
    def expires_in(self) -> Optional[int]:
        value = self.data.get("expires_in")
        return int(value) if value else None

    @property
    def expires_on(self) -> Optional[int]:
        value = self.data.get("expires_on")
        return int(value) if value else None

    @property
    def token_type(self) -> Optional[str]:
        return self.data.get("token_type")

    @property
    def access_token(self) -> Optional[str]:
        return self.data.get("access_token")

    @property
    def refresh_token(self) -> Optional[str]:
        return self.data.get("refresh_token")

    @property
    def id_token(self) -> Optional[str]:
        return self.data.get("id_token")


class BaseTokensStore(ABC):
    """
    Base abstract class for types that can store and restore access tokens and refresh
    tokens in the context of a web request.
    """

    @abstractmethod
    async def store_access_tokens(
        self,
        request: Request,
        response: Response,
        access_token: str,
        refresh_token: Optional[str],
        expires: Optional[datetime] = None,
    ):
        """
        Applies a strategy to store an access token and an optional refresh token for
        the given request and response.
        """

    @abstractmethod
    async def restore_tokens(self, request: Request) -> None:
        """
        Applies a strategy to restore an access token and an optional refresh token for
        the given request.
        """

    async def __call__(self, request: Request, handler):
        await self.restore_tokens(request)
        return await handler(request)


class TokenType(Enum):
    ACCESS_TOKEN = "access_token"
    REFRESH_TOKEN = "refresh_token"


class CookiesTokensStore(BaseTokensStore):
    """
    A class that can store access and refresh tokens in encrypted form in cookies.

    Beware that cookies size can be problematic when storing all information in cookies.
    If this is the case, consider implementing a type of `BaseTokensStore` that uses
    other ways to store and restore tokens (e.g. Redis Cache, etc.).
    """

    def __init__(
        self,
        scheme_name: str = "OpenIDConnect",
        secret_keys: Optional[Sequence[str]] = None,
        serializer: Optional[Serializer] = None,
    ) -> None:
        self._scheme_name: str
        self._access_token_name: str
        self._refresh_token_name: str
        self.scheme_name = scheme_name

        self.serializer = serializer or get_serializer(
            secret_keys, f"{scheme_name}tokens"
        )

    @property
    def scheme_name(self) -> str:
        return self._scheme_name

    @scheme_name.setter
    def scheme_name(self, value: str) -> None:
        self._scheme_name = value
        self._access_token_cookie_name = f"{self._scheme_name}.at".lower()
        self._refresh_token_cookie_name = f"{self._scheme_name}.rt".lower()

    async def store_access_tokens(
        self,
        request: Request,
        response: Response,
        access_token: str,
        refresh_token: Optional[str],
        expires: Optional[datetime] = None,
    ) -> None:
        secure = request.scheme == "https"
        self.set_cookie(
            response,
            self._access_token_cookie_name,
            access_token,
            secure=secure,
            expires=expires,
        )

        if refresh_token:
            self.set_cookie(
                response,
                self._refresh_token_cookie_name,
                refresh_token,
                secure=secure,
                expires=expires,
            )

    async def restore_tokens(self, request: Request) -> None:
        await self._restore_token(
            request, self._access_token_cookie_name, TokenType.ACCESS_TOKEN
        )
        await self._restore_token(
            request, self._refresh_token_cookie_name, TokenType.REFRESH_TOKEN
        )

    def set_cookie(
        self,
        response: Response,
        cookie_name: str,
        data: AnyStr,
        secure: bool,
        expires: Optional[datetime] = None,
    ) -> None:
        value = self.serializer.dumps(data)  # type: ignore
        response.set_cookie(
            Cookie(
                cookie_name,
                ensure_str(value),  # type: ignore
                domain=None,
                path="/",
                http_only=True,
                secure=secure,
                expires=expires,
            )
        )

    async def _restore_token(
        self, request: Request, cookie_name: str, token_type: TokenType
    ) -> None:
        cookie = request.get_cookie(cookie_name)

        if cookie is None:
            pass
        else:
            try:
                value = self.serializer.loads(cookie)
            except BadSignature:
                logger.debug(
                    "Discarding token (%s), invalid signature.",
                    cookie_name,
                )
            else:
                if request.identity:
                    if token_type == TokenType.ACCESS_TOKEN:
                        request.identity.access_token = value
                    elif token_type == TokenType.REFRESH_TOKEN:
                        request.identity.refresh_token = value
        return None


class OpenIDConnectHandler:
    def __init__(
        self,
        settings: OpenIDSettings,
        auth_handler: CookieAuthentication,
        tokens_store: Optional[BaseTokensStore] = None,
        parameters_builder: Optional[ParametersBuilder] = None,
    ) -> None:
        self._settings = settings
        self._configuration: Optional[OpenIDConfiguration] = None
        self._http_handler: HTTPHandler = HTTPHandler()
        self.events = OpenIDConnectEvents(self)
        self.parameters_builder = parameters_builder or ParametersBuilder(settings)
        self.tokens_store = tokens_store
        self._jwt_validator: Optional[JWTValidator] = None
        self._auth_handler = auth_handler

    @property
    def settings(self) -> OpenIDSettings:
        return self._settings

    async def get_jwt_validator(self) -> JWTValidator:
        if self._jwt_validator is None:
            configuration = await self.get_openid_configuration()

            self._jwt_validator = JWTValidator(
                require_kid=True,
                keys_url=configuration.jwks_uri,
                valid_issuers=[configuration.issuer],
                valid_audiences=[self._settings.client_id],
            )
        return self._jwt_validator

    def get_well_known_openid_configuration_url(self) -> str:
        if self._settings.discovery_endpoint:
            return self._settings.discovery_endpoint

        if not self._settings.authority:
            raise OpenIDConnectConfigurationError(
                "Missing `authority` or `discovery_endpoint` in OpenIDSettings. "
                "To fix, define one of the two."
            )

        return (
            self._settings.authority.rstrip("/") + "/.well-known/openid-configuration"
        )

    async def fetch_openid_configuration(self) -> OpenIDConfiguration:
        try:
            data = await self._http_handler.fetch_json(
                self.get_well_known_openid_configuration_url()
            )
        except FailedRequestError as request_error:
            logger.error(
                "Failed to fetch OpenID Connect configuration from the remote endpoint."
                "Inspect the exception details for more details on the cause of the "
                "failure.",
                exc_info=request_error,
            )
            raise OpenIDConnectRequestError(request_error)
        else:
            logger.debug(
                "Fetched OpenID Connect configuration from the remote endpoint."
            )
            return OpenIDConfiguration(data)

    async def get_openid_configuration(self) -> OpenIDConfiguration:
        if self._configuration is None:
            self._configuration = await self.fetch_openid_configuration()

        return self._configuration

    async def redirect_to_sign_in(self, request: Request) -> Response:
        redirect_url = await self.get_redirect_uri(request)
        return redirect(redirect_url)

    async def get_redirect_uri(self, request: Request) -> str:
        openid_conf = await self.get_openid_configuration()
        parameters = self.parameters_builder.build_signin_parameters(request)
        authorization_endpoint = openid_conf.authorization_endpoint
        return authorization_endpoint + "?" + urlencode(parameters)

    async def handle_error(self, request: Request, data: Dict[str, Any]) -> Response:
        """
        Handles an error received from the identity provider.
        """
        query = "?" + urlencode({"error": data.get("error", "unknown")})
        if self._settings.error_redirect_path:
            return redirect(self._settings.error_redirect_path + query)
        return redirect("/" + query)

    async def handle_auth_redirect(self, request: Request) -> Response:
        """
        Handles the redirect after the user interacted with the sign-in page of a remote
        authorization server.
        """
        data = await request.form()

        if data is None or not data:
            return accepted()

        error = data.get("error")

        if error:
            logger.error(
                "Received a post request with error message to the OIDC authorization "
                "callback endpoint: %s",
                error,
            )
            await self.events.on_error.fire(data)
            return await self.handle_error(request, data)

        # the following code validates the state and restores the original path
        # the user was trying to access before being redirected to the OIDC sign-in
        state = data.get("state")
        if isinstance(state, str):
            state = self.parameters_builder.read_state(state)
            redirect_path = state.get("orig_path", "/")
        else:
            redirect_path = "/"

        settings = self._settings

        id_token = data.get("id_token")
        token_response = None

        response = redirect(redirect_path)

        if id_token and not settings.client_secret:
            logger.debug("Successfully obtained an id_token for a user.")
        else:
            code = data.get("code")

            if settings.client_secret and isinstance(code, str):
                # extra call to fetch an access token
                token_response = await self.exchange_token(request, code)

                await self.events.on_tokens_received.fire(token_response)

                if self.tokens_store and token_response.access_token:
                    # ability to store access and refresh tokens - they can be stored
                    # in cookies or for example a Redis Cache
                    await self.tokens_store.store_access_tokens(
                        request,
                        response,
                        token_response.access_token,
                        token_response.refresh_token,
                        expires=None,
                    )

                if not id_token and token_response.id_token:
                    id_token = token_response.id_token
            else:
                raise BadRequest("Expected either an error, an id_token, or a code.")

        if not isinstance(id_token, str):
            # This can happen legitimately if OIDC settings are configured to retrieve
            # only an access token. For example, when using Auth0 and configuring a
            # single scope that does not include openid or profile. In such cases, it is
            # unclear what should be done, especially since access tokens are not stored
            # by default. The user of the library might still being handling the token
            # response using dedicated event, or using a token_store.
            parsed_token = {}
        else:
            parsed_token = await self.validate_id_token(id_token)

            if isinstance(state, dict) and state.get("nonce") != parsed_token.get(
                "nonce"
            ):
                raise OpenIDConnectError("nonce mismatch error")

            await self.events.on_id_token_validated.fire(parsed_token)

        self._auth_handler.set_cookie(
            parsed_token, response, secure=request.scheme == "https"
        )

        return response

    async def exchange_token(self, request: Request, code: str) -> TokenResponse:
        configuration = await self.get_openid_configuration()

        code_grant_parameters = self.parameters_builder.build_code_grant_parameters(
            request, code
        )

        try:
            data = await self._http_handler.post_form(
                configuration.token_endpoint, code_grant_parameters
            )
        except FailedRequestError as request_error:
            logger.error(
                "Failed to exchange an authorization code with an access token. "
                "Inspect the exception details for more details on the cause of the "
                "failure.",
                exc_info=request_error,
            )
            raise OpenIDConnectFailedExchangeError(request_error)
        else:
            logger.debug("Exchanged a code with id_token and access_token for a user.")
            return TokenResponse(data)

    async def refresh_token(self, refresh_token: str) -> TokenResponse:
        configuration = await self.get_openid_configuration()

        refresh_token_parameters = (
            self.parameters_builder.build_refresh_token_parameters(refresh_token)
        )

        try:
            data = await self._http_handler.post_form(
                configuration.token_endpoint, refresh_token_parameters
            )
        except FailedRequestError as request_error:
            logger.error(
                "Failed to exchange a refresh token with an access token. "
                "Inspect the exception details for more details on the cause of the "
                "failure.",
                exc_info=request_error,
            )
            raise OpenIDConnectFailedExchangeError(request_error)
        else:
            logger.debug(
                "Exchanged a refresh token with a fresh access_token for a user."
            )
            return TokenResponse(data)

    async def validate_id_token(self, raw_id_token: str) -> Any:
        jwt_validator = await self.get_jwt_validator()
        try:
            return await jwt_validator.validate_jwt(raw_id_token)
        except (InvalidAccessToken, InvalidTokenError) as ex:
            logger.error("Invalid id_token: %s", str(ex))
            raise Unauthorized(f"Invalid id_token: {ex}")

    async def handle_logout_redirect(self, request: Request) -> Response:
        # TODO: obtain the logout redirect from OIDC.end_session_endpoint
        # and make what is necessary to logout from the remote identity provider.
        # Note that there are differences among providers:
        # e.g. Okta requires the original id_token

        # Auth0 does not provide end_session_endpoint in the discovery endpoint
        # AAD is the simplest scenario because it's a simple redirect to the
        # end_session_endpoint.
        response = redirect(self._settings.post_logout_redirect_path)
        self._auth_handler.unset_cookie(response)
        return response


class ChallengeMiddleware:
    def __init__(
        self, request_handler: Callable[[Request], Awaitable[Response]]
    ) -> None:
        self.request_handler = request_handler

    async def __call__(self, request: Request, handler):
        try:
            return await handler(request)
        except (Unauthorized, UnauthorizedError):
            if (
                request.identity is None
                or not request.identity.is_authenticated()
                and request.method in {"GET", "HEAD"}
            ):
                return await self.request_handler(request)
            raise


def use_openid_connect(
    app: Application,
    settings: OpenIDSettings,
    tokens_store: Optional[BaseTokensStore] = None,
    parameters_builder: Optional[ParametersBuilder] = None,
    is_default: bool = True,
) -> OpenIDConnectHandler:
    """
    Configures an application to use OpenID Connect, integrating with an identity
    provider such as Auth0, Okta, Azure Active Directory.

    Parameters
    ----------
    app : Application
        The application to be configured to handle OpenID Connect.
    settings : OpenIDSettings
        Basic OAuth settings, and other settings to handle the OIDC flow.
    tokens_store : Optional[BaseTokensStore], optional
        If specified, configures the BaseTokensStore used to store access and refresh
        tokens (these are available only when a client_secret is configured), which
        are otherwise not stored., by default None
    parameters_builder : Optional[ParametersBuilder], optional
        ParametersBuilder used to build parameters for OAuth requests, by default None
    is_default : bool, optional
        If true, the application is configured to automatically redirect
        not-authenticated users to the sign-in endpoint, by default True

    Returns
    -------
    OpenIDConnectHandler
        Instance of a class that handles the OIDC integration.
    """
    scheme_name = settings.scheme_name or "OpenIDConnect"
    auth_handler = CookieAuthentication(
        cookie_name=scheme_name.lower(), auth_scheme=scheme_name
    )
    app.use_authentication().add(auth_handler)

    handler = OpenIDConnectHandler(
        settings,
        tokens_store=tokens_store,
        parameters_builder=parameters_builder,
        auth_handler=auth_handler,
    )

    @allow_anonymous()
    @app.router.get(settings.entry_path)
    async def redirect_to_sign_in(request: Request):
        return await handler.redirect_to_sign_in(request)

    @allow_anonymous()
    @app.router.post(settings.callback_path)
    async def handle_auth_redirect(request: Request):
        return await handler.handle_auth_redirect(request)

    @app.router.get(settings.logout_path)
    async def redirect_to_logout(request: Request):
        return await handler.handle_logout_redirect(request)

    if is_default:

        @app.on_middlewares_configuration
        def insert_challenge_middleware(app):
            app.middlewares.insert(0, ChallengeMiddleware(handler.redirect_to_sign_in))

    if tokens_store:
        app.middlewares.append(tokens_store)

    return handler



import html
import traceback

from blacksheep.contents import HTMLContent
from blacksheep.messages import Request, Response
from blacksheep.server.asgi import get_request_url
from blacksheep.server.resources import get_resource_file_content


def _load_error_page_template() -> str:
    error_css = get_resource_file_content("error.css")
    error_template = get_resource_file_content("error.html")
    assert "/*STYLES*/" in error_template

    # since later it is used in format_map...
    error_css = error_css.replace("{", "{{").replace("}", "}}")
    return error_template.replace("/*STYLES*/", error_css)


class ServerErrorDetailsHandler:
    """
    This class is responsible of producing a detailed response when the Application is
    configured to show error details to the client, and an unhandled exception happens.
    """

    def __init__(self) -> None:
        self._error_page_template = _load_error_page_template()

    def produce_response(self, request: Request, exc: Exception) -> Response:
        tb = traceback.format_exception(exc.__class__, exc, exc.__traceback__)
        info = ""
        for item in tb:
            info += f"<li><pre>{html.escape(item)}</pre></li>"

        content = HTMLContent(
            self._error_page_template.format_map(
                {
                    "info": info,
                    "exctype": exc.__class__.__name__,
                    "excmessage": str(exc),
                    "method": request.method,
                    "path": request.url.value.decode(),
                    "full_url": get_request_url(request),
                }
            )
        )

        return Response(500, content=content)
